"""
(c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import abc
import json
import multiprocessing
from multiprocessing import queues
import shutil
import tempfile
import time

from oslo_log import log
import six
# PyCharm will not recognize queue. Puts red squiggle line under it. That's OK.
from six.moves import queue

from freezer.exceptions import engine as engine_exceptions
from freezer.storage import base
from freezer.utils import streaming
from freezer.utils import utils

LOG = log.getLogger(__name__)


@six.add_metaclass(abc.ABCMeta)
class BackupEngine(object):
    """
    The main part of making a backup and making a restore is the mechanism of
    implementing it. For a long time Freezer had only one mechanism of
    doing it - invoking gnutar and it was heavy hard-coded.

    Currently we are going to support many different approaches.
    One of them is rsync. Having many different implementations requires to
    have an abstraction level

    This class is an abstraction over all implementations.

    Workflow:
    1) invoke backup
        1.1) try to download metadata for incremental
        1.2) create a dataflow between backup_stream and storage.write_backup
            Backup_stream is producer of data, for tar backup
            it creates a gnutar subprocess and start to read data from stdout
            Storage write_backup is consumer of data, it creates a thread
            that store data in storage.
            Both streams communicate in non-blocking mode
        1.3) invoke post_backup - now it uploads metadata file
    2) restore backup
        2.1) define all incremental backups
        2.2) for each incremental backup create a dataflow between
            storage.read_backup and restore_stream
            Read_backup is data producer, it reads data chunk by chunk from
            the specified storage and pushes the chunks into a queue.
            Restore stream is a consumer, that is actually does restore (for
            tar it is a thread that creates gnutar subprocess and feeds chunks
            to stdin of this thread.

    :type storage: freezer.storage.base.Storage
    """

    def __init__(self, storage):
        """
        :type storage: freezer.storage.base.Storage
        :param storage:
        :return:
        """
        self.storage = storage

    @abc.abstractproperty
    def name(self):
        """
        :rtype: str
        :return: Engine name
        """
        pass

    def backup_stream(self, backup_resource, rich_queue, manifest_path):
        """
        :param rich_queue:
        :type rich_queue: freezer.streaming.RichQueue
        :type manifest_path: list[str]
        :param manifest_path:
        :return:
        """
        rich_queue.put_messages(self.backup_data(backup_resource,
                                                 manifest_path))

    def backup(self, backup_resource, hostname_backup_name, no_incremental,
               max_level, always_level, restart_always_level, queue_size=2):
        """
        Here we now location of all interesting artifacts like metadata
        Should return stream for storing data.
        :return: stream
        """
        prev_backup = self.storage.previous_backup(
            engine=self,
            hostname_backup_name=hostname_backup_name,
            no_incremental=no_incremental,
            max_level=max_level,
            always_level=always_level,
            restart_always_level=restart_always_level
        )

        try:
            tmpdir = tempfile.mkdtemp()
        except Exception:
            LOG.error("Unable to create a tmp directory")
            raise

        try:
            engine_meta = utils.path_join(tmpdir, "engine_meta")
            freezer_meta = utils.path_join(tmpdir, "freezer_meta")
            if prev_backup:
                prev_backup.storage.get_file(prev_backup.engine_metadata_path,
                                             engine_meta)
            timestamp = utils.DateTime.now().timestamp
            level_zero_timestamp = (prev_backup.level_zero_timestamp
                                    if prev_backup else timestamp)
            backup = base.Backup(
                engine=self,
                hostname_backup_name=hostname_backup_name,
                level_zero_timestamp=level_zero_timestamp,
                timestamp=timestamp,
                level=(prev_backup.level + 1 if prev_backup else 0)
            )

            input_queue = streaming.RichQueue(queue_size)
            read_except_queue = queue.Queue()
            write_except_queue = queue.Queue()

            read_stream = streaming.QueuedThread(
                self.backup_stream,
                input_queue,
                read_except_queue,
                kwargs={"backup_resource": backup_resource,
                        "manifest_path": engine_meta})

            write_stream = streaming.QueuedThread(
                self.storage.write_backup,
                input_queue,
                write_except_queue,
                kwargs={"backup": backup})

            read_stream.daemon = True
            write_stream.daemon = True
            read_stream.start()
            write_stream.start()
            read_stream.join()
            write_stream.join()

            # queue handling is different from SimpleQueue handling.
            def handle_except_queue(except_queue):
                if not except_queue.empty():
                    while not except_queue.empty():
                        e = except_queue.get_nowait()
                        LOG.critical('Engine error: {0}'.format(e))
                    return True
                else:
                    return False

            got_exception = None
            got_exception = (handle_except_queue(read_except_queue) or
                             got_exception)
            got_exception = (handle_except_queue(write_except_queue) or
                             got_exception)

            if got_exception:
                raise engine_exceptions.EngineException(
                    "Engine error. Failed to backup.")

            with open(freezer_meta, mode='wb') as b_file:
                b_file.write(json.dumps(self.metadata(backup_resource)))
            self.storage.put_metadata(engine_meta, freezer_meta, backup)
        finally:
            shutil.rmtree(tmpdir)

    def read_blocks(self, backup, write_pipe, read_pipe, except_queue):
        # Close the read pipe in this child as it is unneeded
        # and download the objects from swift in chunks. The
        # Chunk size is set by RESP_CHUNK_SIZE and sent to che write
        # pipe

        try:

            read_pipe.close()
            for block in backup.storage.backup_blocks(backup):
                write_pipe.send_bytes(block)

            # Closing the pipe after checking no data
            # is still available in the pipe.
            while True:
                if not write_pipe.poll():
                    write_pipe.close()
                    break
                time.sleep(1)

        except IOError:
            pass

        except Exception as e:
            except_queue.put(e)
            raise

    def restore(self, hostname_backup_name, restore_resource,
                overwrite,
                recent_to_date,
                backup_media=None):
        """

        :param hostname_backup_name:
        :param restore_path:
        :param overwrite:
        :param recent_to_date:
        """
        if backup_media == 'fs':
            LOG.info("Creating restore path: {0}".format(restore_resource))
            # if restore path can't be created this function will raise
            # exception
            utils.create_dir_tree(restore_resource)
            if not overwrite and not utils.is_empty_dir(restore_resource):
                raise Exception(
                    "Restore dir is not empty. "
                    "Please use --overwrite or provide different path "
                    "or remove the content of {}".format(restore_resource))

            LOG.info("Restore path creation completed")

        backups = self.storage.get_latest_level_zero_increments(
            engine=self,
            hostname_backup_name=hostname_backup_name,
            recent_to_date=recent_to_date)

        max_level = max(backups.keys())

        # Use SimpleQueue because Queue does not work on Mac OS X.
        read_except_queue = queues.SimpleQueue()
        LOG.info("Restoring backup {0}".format(hostname_backup_name))
        for level in range(0, max_level + 1):
            LOG.info("Restoring from level {0}".format(level))
            backup = backups[level]
            read_pipe, write_pipe = multiprocessing.Pipe()
            process_stream = multiprocessing.Process(
                target=self.read_blocks,
                args=(backup, write_pipe, read_pipe, read_except_queue))

            process_stream.daemon = True
            process_stream.start()
            write_pipe.close()

            # Start the tar pipe consumer process

            # Use SimpleQueue because Queue does not work on Mac OS X.
            write_except_queue = queues.SimpleQueue()

            engine_stream = multiprocessing.Process(
                target=self.restore_level,
                args=(restore_resource, read_pipe, backup, write_except_queue))

            engine_stream.daemon = True
            engine_stream.start()

            read_pipe.close()
            write_pipe.close()
            process_stream.join()
            engine_stream.join()

            # SimpleQueue handling is different from queue handling.
            def handle_except_SimpleQueue(except_queue):
                if not except_queue.empty():
                    while not except_queue.empty():
                        e = except_queue.get()
                        LOG.exception('Engine error: {0}'.format(e))
                    return True
                else:
                    return False

            got_exception = None
            got_exception = (handle_except_SimpleQueue(read_except_queue) or
                             got_exception)
            got_exception = (handle_except_SimpleQueue(write_except_queue) or
                             got_exception)

            if engine_stream.exitcode or got_exception:
                raise engine_exceptions.EngineException(
                    "Engine error. Failed to restore.")

        LOG.info(
            'Restore completed successfully for backup name '
            '{0}'.format(hostname_backup_name))

    @abc.abstractmethod
    def restore_level(self, restore_path, read_pipe, backup, except_queue):
        pass

    @abc.abstractmethod
    def backup_data(self, backup_path, manifest_path):
        """
        :param backup_path:
        :param manifest_path:
        :return:
        """
        pass

    @abc.abstractmethod
    def metadata(self, backup_resource):
        pass



"""
Copyright 2015 Hewlett-Packard

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import json
import os
import signal
import socket

from freezerclient import exceptions
from oslo_log import log
from oslo_utils import uuidutils
import psutil


LOG = log.getLogger(__name__)

CONFIG_FILE_EXT = '.conf'


def do_register(client, args=None):
    if client:
        client_info = {
            "client_id": client.client_id,
            "hostname": socket.gethostname()
        }
        try:
            client.clients.create(client_info)
        except exceptions.ApiClientException as e:
            if e.status_code == 409:
                print("Client already registered")
            return 73  # os.EX_CANTCREAT
        return 0  # os.EX_OK


def find_config_files(path):
    expanded_path = os.path.expanduser(path)
    if os.path.isfile(expanded_path):
        return [expanded_path]
    file_list = []
    if os.path.isdir(expanded_path):
        for fname in os.walk(expanded_path).next()[2]:
            if CONFIG_FILE_EXT.upper() == os.path.splitext(fname)[1].upper():
                file_list.append('{0}/{1}'.format(expanded_path, fname))
        return file_list
    raise Exception("unable to find job files at the provided path "
                    "{0}".format(path))


def load_doc_from_json_file(fname, debug=False):
    with open(fname, 'rb') as fd:
        try:
            doc = json.load(fd)
        except Exception as e:
            raise Exception("Unable to load conf file. {0}".format(e))
        if debug:
            print("File {0} loaded: ".format(fname))
        return doc


def save_doc_to_json_file(doc, fname, debug=False):
    with open(fname, 'w') as fd:
        json.dump(doc, fd, indent=4)
    if debug:
        print('Saved doc to file: {0}'.format(fname))


def get_jobs_from_disk(path):
    job_doc_list = [
        load_doc_from_json_file(f) for f in find_config_files(path)]
    for job_doc in job_doc_list:
        if job_doc:
            job_doc['job_id'] = job_doc.get('job_id', uuidutils.generate_uuid(
                dashed=False))
    return [x for x in job_doc_list if x]


def save_jobs_to_disk(job_doc_list, path):
    for doc in job_doc_list:
        fname = os.path.normpath('{0}/job_{1}.conf'.
                                 format(path, doc['job_id']))
        save_doc_to_json_file(doc, fname)


def get_active_jobs_from_api(client):
    # might raise
    search = {"match_not": [{"status": "completed"}]}
    job_list, offset = [], 0
    while True:
        jobs = client.jobs.list(limit=10, offset=offset, search=search)
        job_list.extend(jobs)
        if len(jobs) < 10:
            break
        offset += len(jobs)
    return job_list


def terminate_subprocess(pid, name):
    try:
        process = psutil.Process(pid)
        if process.name.startswith(name):
            os.kill(pid, signal.SIGTERM)
        else:
            LOG.warning('The name {} does not match the pid {}'.format(
                name, pid))
    except Exception:
        LOG.debug('Process {} does not exists anymore'.format(pid))



# Copyright 2015 Hewlett-Packard
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========================================================================

from copy import copy
import json
import os
import unittest

from oslo_utils import uuidutils

from freezer.tests.integration import common


class TestSimpleExecution(common.TestFS):
    def test_freezerc_executes(self):
        result = common.execute_freezerc({'version': ''})
        self.assertIsNotNone(result)

    def test_freezerc_fails_with_wrong_params(self):
        result = common.execute_freezerc({'blabla': ''}, must_fail=True,
                                         merge_stderr=True)
        self.assertIn('unrecognized arguments', result)


class TestBackupFSLocalstorage(common.TestFS):
    def test_trees(self):
        self.assertTreesMatch()
        self.source_tree.add_random_data()
        self.assertTreesMatchNot()

    def test_backup_single_level(self):
        """
        - use the default source and destination trees in /tmp
        (see common.TestFS)
        - use temporary directory for backup storage
        - add some random data
        - check that trees don't match anymore
        - execute backup of source tree
        - execute restore into destination tree
        - check that source and destination trees match

        :return: non on success
        """
        self.source_tree.add_random_data()
        self.assertTreesMatchNot()

        with common.Temp_Tree() as storage_dir:
            backup_args = {
                'action': 'backup',
                'mode': 'fs',
                'path_to_backup': self.source_tree.path,
                'container': storage_dir.path,
                'storage': 'local',
                'max_level': '6',
                'max_segment_size': '67108864',
                'backup_name': uuidutils.generate_uuid(dashed=False)
            }

            restore_args = {
                'action': 'restore',
                'restore_abs_path': self.dest_tree.path,
                'backup_name': copy(backup_args['backup_name']),
                'storage': 'local',
                'container': storage_dir.path
            }
            result = common.execute_freezerc(backup_args)
            self.assertIsNotNone(result)
            result = common.execute_freezerc(restore_args)
            self.assertIsNotNone(result)
        self.assertTreesMatch()

    def test_backup_preexisting_dir(self):
        """
        Use external pre-defined directory for tests.
        If directory does not exist, then skip

        Restore to temporary folder (removed on exit)
        :return:
        """
        workdir = os.path.expanduser('~/test_dir')
        if not os.path.isdir(workdir):
            return
        self.source_tree = common.Temp_Tree(dir='/work', create=False)

        with common.Temp_Tree() as storage_dir:
            backup_args = {
                'action': 'backup',
                'mode': 'fs',
                'path_to_backup': self.source_tree.path,
                'container': storage_dir.path,
                'storage': 'local',
                'max_level': '6',
                'max_segment_size': '67108864',
                'backup_name': uuidutils.generate_uuid(dashed=False)
            }

            restore_args = {
                'action': 'restore',
                'restore_abs_path': self.dest_tree.path,
                'backup_name': copy(backup_args['backup_name']),
                'storage': 'local',
                'container': storage_dir.path
            }
            result = common.execute_freezerc(backup_args)
            self.assertIsNotNone(result)
            result = common.execute_freezerc(restore_args)
            self.assertIsNotNone(result)
        self.assertTreesMatch()

    @unittest.skipIf(not common.TestFS.use_lvm, "No LVM support")
    def test_backup_local_storage_use_lvm_snapshot_and_path_to_backup(self):
        self.source_tree.add_random_data()
        self.assertTreesMatchNot()

        backup_name = uuidutils.generate_uuid(dashed=False)
        path_to_backup = self.source_tree.path
        lvm_snapsize = '50M'
        lvm_snapname = 'freezer-snap_{0}'.format(backup_name)
        lvm_dirmount = '/var/freezer/freezer-{0}'.format(backup_name)

        with common.Temp_Tree() as storage_dir:
            backup_args = {
                'action': 'backup',
                'mode': 'fs',
                'path_to_backup': path_to_backup,
                'snapshot': '',
                'lvm_dirmount': lvm_dirmount,
                'lvm_snapsize': lvm_snapsize,
                'lvm_snapname': lvm_snapname,
                'container': storage_dir.path,
                'storage': 'local',
                'max_level': '6',
                'max_segment_size': '67108864',
                'backup_name': backup_name
            }
            restore_args = {
                'action': 'restore',
                'restore_abs_path': self.dest_tree.path,
                'backup_name': copy(backup_args['backup_name']),
                'storage': 'local',
                'container': storage_dir.path
            }

            result = common.execute_freezerc(backup_args)
            self.assertIsNotNone(result)
            result = common.execute_freezerc(restore_args)
            self.assertIsNotNone(result)
        self.assertTreesMatch()


class TestBackupSSH(common.TestFS):
    """
    Tests are executed if the following env vars are defined:
    - FREEZER_TEST_SSH_KEY
    - FREEZER_TEST_SSH_USERNAME
    - FREEZER_TEST_SSH_HOST
    - FREEZER_TEST_CONTAINER
    (directory on the remote machine used to store backups)
    """

    @unittest.skipIf(not common.TestFS.use_ssh,
                     "Cannot test with ssh, please provide"
                     "'FREEZER_TEST_SSH_KEY,'"
                     "'FREEZER_TEST_SSH_USERNAME',"
                     "'FREEZER_TEST_SSH_HOST',"
                     "'FREEZER_TEST_CONTAINER'")
    def test_backup_ssh(self):
        self.source_tree.add_random_data()
        self.assertTreesMatchNot()

        backup_args = {
            'action': 'backup',
            'mode': 'fs',
            'path_to_backup': self.source_tree.path,
            'max_level': '6',
            'max_segment_size': '67108864',
            'backup_name': uuidutils.generate_uuid(dashed=False),
            'storage': 'ssh',
            'container': self.container,
            'ssh_key': self.ssh_key,
            'ssh_username': self.ssh_username,
            'ssh_host': self.ssh_host,
            'metadata_out': '-'
        }
        restore_args = {
            'action': 'restore',
            'restore_abs_path': self.dest_tree.path,
            'backup_name': copy(backup_args['backup_name']),
            'storage': 'ssh',
            'container': self.container,
            'ssh_key': self.ssh_key,
            'ssh_username': self.ssh_username,
            'ssh_host': self.ssh_host
        }

        result = common.execute_freezerc(backup_args)
        self.assertIsNotNone(result)
        result = json.loads(result)
        sub_path = '_'.join([result['hostname'], result['backup_name']])
        # It may be reasonable to insert a check of the files in the
        # storage directory
        # file_list = self.get_file_list_ssh(sub_path)

        self.assertIn('backup_name', result)
        self.assertEqual(result['backup_name'], backup_args['backup_name'])
        self.assertIn('container', result)
        self.assertEqual(result['container'], self.container)

        result = common.execute_freezerc(restore_args)
        self.assertIsNotNone(result)
        self.assertTreesMatch()

        self.remove_ssh_directory(sub_path)

    @unittest.skipIf(not common.TestFS.use_ssh,
                     "Cannot test with ssh, please provide"
                     "'FREEZER_TEST_SSH_KEY,'"
                     "'FREEZER_TEST_SSH_USERNAME',"
                     "'FREEZER_TEST_SSH_HOST',"
                     "'FREEZER_TEST_CONTAINER'")
    def test_backup_ssh_incremental(self):
        self.source_tree.add_random_data()
        self.assertTreesMatchNot()

        backup_args = {
            'action': 'backup',
            'mode': 'fs',
            'path_to_backup': self.source_tree.path,
            'max_level': '6',
            'max_segment_size': '67108864',
            'backup_name': uuidutils.generate_uuid(dashed=False),
            'storage': 'ssh',
            'container': self.container,
            'ssh_key': self.ssh_key,
            'ssh_username': self.ssh_username,
            'ssh_host': self.ssh_host,
            'metadata_out': '-'
        }
        restore_args = {
            'action': 'restore',
            'restore_abs_path': self.dest_tree.path,
            'backup_name': copy(backup_args['backup_name']),
            'storage': 'ssh',
            'container': self.container,
            'ssh_key': self.ssh_key,
            'ssh_username': self.ssh_username,
            'ssh_host': self.ssh_host
        }
        result = common.execute_freezerc(backup_args)
        self.assertIsNotNone(result)

        result = json.loads(result)
        sub_path = '_'.join([result['hostname'], result['backup_name']])
        # It may be reasonable to insert a check of the files in the
        # storage directory
        # file_list = self.get_file_list_ssh(sub_path)

        result = common.execute_freezerc(restore_args)
        self.assertIsNotNone(result)
        self.assertTreesMatch()

        # -- many levels
        self.do_backup_and_restore_with_check(backup_args, restore_args)
        self.do_backup_and_restore_with_check(backup_args, restore_args)
        self.do_backup_and_restore_with_check(backup_args, restore_args)
        self.do_backup_and_restore_with_check(backup_args, restore_args)
        self.do_backup_and_restore_with_check(backup_args, restore_args)
        self.do_backup_and_restore_with_check(backup_args, restore_args)
        self.do_backup_and_restore_with_check(backup_args, restore_args)

        self.remove_ssh_directory(sub_path)

    @unittest.skipIf(not common.TestFS.use_ssh,
                     "Cannot test with ssh, please provide"
                     "'FREEZER_TEST_SSH_KEY,'"
                     "'FREEZER_TEST_SSH_USERNAME',"
                     "'FREEZER_TEST_SSH_HOST',"
                     "'FREEZER_TEST_CONTAINER'")
    @unittest.skipIf(not common.TestFS.use_lvm, "No LVM support")
    def test_backup_ssh_incremental_with_lvm(self):
        self.source_tree.add_random_data()
        self.assertTreesMatchNot()

        backup_name = uuidutils.generate_uuid(dashed=False)
        path_to_backup = self.source_tree.path
        lvm_snapsize = '1G'
        lvm_snapname = 'freezer-snap_{0}'.format(backup_name)
        lvm_dirmount = '/var/freezer/freezer-{0}'.format(backup_name)

        backup_args = {
            'action': 'backup',
            'mode': 'fs',
            'path_to_backup': path_to_backup,
            'lvm_dirmount': lvm_dirmount,
            'lvm_snapsize': lvm_snapsize,
            'lvm_snapname': lvm_snapname,
            'backup_name': backup_name,
            'max_level': '6',
            'max_segment_size': '67108864',
            'storage': 'ssh',
            'container': self.container,
            'ssh_key': self.ssh_key,
            'ssh_username': self.ssh_username,
            'ssh_host': self.ssh_host
        }
        restore_args = {
            'action': 'restore',
            'restore_abs_path': self.dest_tree.path,
            'backup_name': copy(backup_args['backup_name']),
            'storage': 'ssh',
            'container': self.container,
            'ssh_key': self.ssh_key,
            'ssh_username': self.ssh_username,
            'ssh_host': self.ssh_host
        }
        result = common.execute_freezerc(backup_args)
        self.assertIsNotNone(result)
        result = common.execute_freezerc(restore_args)
        self.assertIsNotNone(result)
        self.assertTreesMatch()

        # -- level 1
        self.source_tree.add_random_data()
        self.assertTreesMatchNot()
        result = common.execute_freezerc(backup_args)
        self.assertIsNotNone(result)
        result = common.execute_freezerc(restore_args)
        self.assertIsNotNone(result)
        self.assertTreesMatch()

        # -- level 2
        self.source_tree.add_random_data()
        self.assertTreesMatchNot()
        result = common.execute_freezerc(backup_args)
        self.assertIsNotNone(result)
        result = common.execute_freezerc(restore_args)
        self.assertIsNotNone(result)
        self.assertTreesMatch()


class TestBackupUsingSwiftStorage(common.TestFS):
    """
    Tests are executed if the following env vars are defined:

    - FREEZER_TEST_OS_TENANT_NAME
    - FREEZER_TEST_OS_USERNAME
    - FREEZER_TEST_OS_REGION_NAME
    - FREEZER_TEST_OS_PASSWORD
    - FREEZER_TEST_OS_AUTH_URL
    """

    @unittest.skipIf(not common.TestFS.use_os,
                     "Cannot test with swift, please provide"
                     "'FREEZER_TEST_OS_TENANT_NAME',"
                     "'FREEZER_TEST_OS_USERNAME',"
                     "'FREEZER_TEST_OS_REGION_NAME',"
                     "'FREEZER_TEST_OS_PASSWORD',"
                     "'FREEZER_TEST_OS_AUTH_URL'")
    def test_backup_os_simple(self):
        self.source_tree.add_random_data()
        self.assertTreesMatchNot()

        backup_args = {
            'action': 'backup',
            'mode': 'fs',
            'path_to_backup': self.source_tree.path,
            'max_level': '6',
            'max_segment_size': '67108864',
            'backup_name': uuidutils.generate_uuid(dashed=False),
            'storage': 'swift',
            'container': 'freezer_test_backups_{0}'.format(
                uuidutils.generate_uuid(dashed=False)),
            'metadata_out': '-'
        }
        restore_args = {
            'action': 'restore',
            'restore_abs_path': self.dest_tree.path,
            'backup_name': copy(backup_args['backup_name']),
            'storage': 'swift',
            'container': copy(backup_args['container']),
        }
        remove_args = {
            'action': 'admin',
            'remove_older_than': 0,
            'backup_name': copy(backup_args['backup_name']),
            'storage': 'swift',
            'container': copy(backup_args['container']),
        }
        # --- backup
        result = common.execute_freezerc(backup_args)
        self.assertIsNotNone(result)
        result = json.loads(result)
        self.assertIn('backup_name', result)
        self.assertEqual(result['backup_name'], backup_args['backup_name'])
        self.assertIn('container', result)
        self.assertEqual(result['container'], backup_args['container'])

        # It may be reasonable to insert a check of the files in the
        # swift container
        # file_list = self.get_file_list_openstack(result['container'])

        # --- restore
        result = common.execute_freezerc(restore_args)
        self.assertIsNotNone(result)
        self.assertTreesMatch()

        # --- remove backups and container
        result = common.execute_freezerc(remove_args)
        self.assertIsNotNone(result)

        result = self.remove_swift_container(backup_args['container'])
        self.assertIsNotNone(result)

    @unittest.skipIf(not common.TestFS.use_os,
                     "Cannot test with swift, please provide"
                     "'FREEZER_TEST_OS_TENANT_NAME',"
                     "'FREEZER_TEST_OS_USERNAME',"
                     "'FREEZER_TEST_OS_REGION_NAME',"
                     "'FREEZER_TEST_OS_PASSWORD',"
                     "'FREEZER_TEST_OS_AUTH_URL'")
    def test_backup_os_simple_with_bzip2(self):
        self.source_tree.add_random_data()
        self.assertTreesMatchNot()

        backup_args = {
            'action': 'backup',
            'mode': 'fs',
            'compression': 'bzip2',
            'path_to_backup': self.source_tree.path,
            'max_level': '6',
            'max_segment_size': '67108864',
            'backup_name': uuidutils.generate_uuid(dashed=False),
            'storage': 'swift',
            'container': 'freezer_test_backups_{0}'.format(
                uuidutils.generate_uuid(dashed=False)),
            'metadata_out': '-'
        }
        restore_args = {
            'action': 'restore',
            'compression': 'bzip2',
            'restore_abs_path': self.dest_tree.path,
            'backup_name': copy(backup_args['backup_name']),
            'storage': 'swift',
            'container': copy(backup_args['container']),
        }
        remove_args = {
            'action': 'admin',
            'remove_older_than': 0,
            'backup_name': copy(backup_args['backup_name']),
            'storage': 'swift',
            'container': copy(backup_args['container']),
        }
        # --- backup
        result = common.execute_freezerc(backup_args)
        self.assertIsNotNone(result)
        result = json.loads(result)
        self.assertIn('backup_name', result)
        self.assertEqual(result['backup_name'], backup_args['backup_name'])
        self.assertIn('container', result)
        self.assertEqual(result['container'], backup_args['container'])

        # It may be reasonable to insert a check of the files in the
        # swift container
        # file_list = self.get_file_list_openstack(result['container'])

        # --- restore
        result = common.execute_freezerc(restore_args)
        self.assertIsNotNone(result)
        self.assertTreesMatch()

        # --- remove backups and container
        result = common.execute_freezerc(remove_args)
        self.assertIsNotNone(result)

        result = self.remove_swift_container(backup_args['container'])
        self.assertIsNotNone(result)

    @unittest.skipIf(not common.TestFS.use_os,
                     "Cannot test with swift, please provide"
                     "'FREEZER_TEST_OS_TENANT_NAME',"
                     "'FREEZER_TEST_OS_USERNAME',"
                     "'FREEZER_TEST_OS_REGION_NAME',"
                     "'FREEZER_TEST_OS_PASSWORD',"
                     "'FREEZER_TEST_OS_AUTH_URL'")
    @unittest.skipIf(not common.TestFS.use_lvm, "No LVM support")
    @unittest.skipIf(not os.path.isdir('/var/lib/mysql'),
                     "No path /var/lib/mysql")
    def test_backup_swift_mysql(self):
        self.source_tree = common.Temp_Tree(dir='/var/lib/mysql', create=False)

        backup_name = uuidutils.generate_uuid(dashed=False)
        lvm_snapsize = '1G'
        lvm_snapname = 'freezer-snap_{0}'.format(backup_name)
        lvm_dirmount = '/var/freezer/freezer-{0}'.format(backup_name)

        backup_args = {
            'action': 'backup',
            'mode': 'mysql',
            'mysql_conf': '/etc/mysql/debian.cnf',
            'path_to_backup': self.source_tree.path,
            'snapshot': '',
            'lvm_dirmount': lvm_dirmount,
            'lvm_snapsize': lvm_snapsize,
            'lvm_snapname': lvm_snapname,
            'container': 'freezer_test_container_{0}'.format(backup_name),
            'storage': 'swift',
            'max_level': '6',
            'max_segment_size': '67108864',
            'backup_name': backup_name
        }
        restore_args = {
            'action': 'restore',
            'restore_abs_path': self.dest_tree.path,
            'backup_name': copy(backup_args['backup_name']),
            'storage': 'swift',
            'container': copy(backup_args['container'])
        }

        result = common.execute_freezerc(backup_args)
        self.assertIsNotNone(result)
        result = common.execute_freezerc(restore_args)
        self.assertIsNotNone(result)
        # we cannot test if trees as a running mysql instance
        # will modify the files

    @unittest.skipIf(not common.TestFS.use_os,
                     "Cannot test with swift, please provide"
                     "'FREEZER_TEST_OS_TENANT_NAME',"
                     "'FREEZER_TEST_OS_USERNAME',"
                     "'FREEZER_TEST_OS_REGION_NAME',"
                     "'FREEZER_TEST_OS_PASSWORD',"
                     "'FREEZER_TEST_OS_AUTH_URL'")
    def test_backup_os_simple_with_bandwidth_limit(self):
        self.source_tree.add_random_data()
        self.assertTreesMatchNot()

        backup_args = {
            'action': 'backup',
            'mode': 'fs',
            'path_to_backup': self.source_tree.path,
            'max_level': '6',
            'upload_limit': '1M',
            'download_limit': '1M',
            'max_segment_size': '67108864',
            'backup_name': uuidutils.generate_uuid(dashed=False),
            'storage': 'swift',
            'container': 'freezer_test_backups_{0}'.format(
                uuidutils.generate_uuid(dashed=False)),
            'metadata_out': '-'
        }
        restore_args = {
            'action': 'restore',
            'restore_abs_path': self.dest_tree.path,
            'backup_name': copy(backup_args['backup_name']),
            'storage': 'swift',
            'container': copy(backup_args['container']),
        }
        remove_args = {
            'action': 'admin',
            'remove_older_than': 0,
            'backup_name': copy(backup_args['backup_name']),
            'storage': 'swift',
            'container': copy(backup_args['container']),
        }
        # --- backup
        result = common.execute_freezerc(backup_args)
        self.assertIsNotNone(result)
        result = json.loads(result)
        self.assertIn('backup_name', result)
        self.assertEqual(result['backup_name'], backup_args['backup_name'])
        self.assertIn('container', result)
        self.assertEqual(result['container'], backup_args['container'])

        # It may be reasonable to insert a check of the files in the
        # swift container
        # file_list = self.get_file_list_openstack(result['container'])

        # --- restore
        result = common.execute_freezerc(restore_args)
        self.assertIsNotNone(result)
        self.assertTreesMatch()

        # --- remove backups and container
        result = common.execute_freezerc(remove_args)
        self.assertIsNotNone(result)

        result = self.remove_swift_container(backup_args['container'])
        self.assertIsNotNone(result)



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from datetime import datetime
from datetime import timedelta
import json
import os
import subprocess

from time import mktime

from tempest import test

from freezer.tests.integration.common import Temp_Tree


def resolve_paths(metadata):
    """Find all paths associated with a particular backup

    freezer-agent stores all backups in the timestamped sub-directory of the
    first backup created with a particular (name, hostname) pair, so it isn't
    possible to guess the true location of backup data. This function searches
    the backup container to find both the true parent directory and a list of
    all files associated with the given metadata.

    :param metadata: the metadata associated with the backup to resolve
    :return: a tuple containing the parent directory and a list of associated
             files, or (None, None)
    """
    base_name = '{}_{}'.format(metadata['hostname'], metadata['backup_name'])
    expected_name = '{}_{}'.format(base_name, metadata['time_stamp'])

    backup_base_path = os.path.join(metadata['container'], base_name)
    for timestamp in os.listdir(backup_base_path):
        timestamp_abs = os.path.join(backup_base_path, timestamp)

        matching = filter(lambda p: expected_name in p,
                          os.listdir(timestamp_abs))
        if matching:
            return timestamp_abs, matching

    return None, None


def mutate_timestamp(metadata, days_old):
    """Alter all timestamps of an existing backup

    Since there's no proper way to assign a timestamp to a backup, this method
    takes an existing backup and modifies all associated timestamps to make it
    otherwise indistinguishable from a backup actually created in the past.

    :param metadata: the metadata associated with the backup to mutate
    :param days_old: the age (i.e. days before now) that should be set
    """
    date = datetime.now() - timedelta(days=days_old)
    old_time_stamp = metadata['time_stamp']
    new_time_stamp = int(mktime(date.timetuple()))

    parent_dir, files = resolve_paths(metadata)
    if os.path.basename(parent_dir) == str(old_time_stamp):
        # rename the parent dir, but only if it was created for this
        # backup (the dir may contain other backups with different
        # timestamps that we shouldn't touch)
        new_path = os.path.join(os.path.dirname(parent_dir),
                                str(new_time_stamp))
        os.rename(parent_dir, new_path)
        parent_dir = new_path

    # rename each file associated with the backup, since each filename
    # contains the timestamp as well
    for old_file in files:
        new_file = old_file.replace(str(old_time_stamp), str(new_time_stamp))
        os.rename(os.path.join(parent_dir, old_file),
                  os.path.join(parent_dir, new_file))

    # update the metadata before saving to keep things consistent
    metadata['time_stamp'] = new_time_stamp


def load_metadata(path):
    """Given a metadata path, return a dict containing parsed values.

    :param path: the path to load
    :return: a metadata dict
    """
    with open(path, 'r') as f:
        return json.load(f)


def save_metadata(metadata, path):
    """Write the given metadata object to the provided path.

    :param metadata: the metadata dict to write
    :param path: the path at which to write the metadata
    """
    with open(path, 'w') as f:
        json.dump(metadata, f)


class BaseFreezerTest(test.BaseTestCase):
    credentials = ['primary']

    def __init__(self, *args, **kwargs):

        super(BaseFreezerTest, self).__init__(*args, **kwargs)

    # noinspection PyAttributeOutsideInit
    def setUp(self):
        super(BaseFreezerTest, self).setUp()

        self.storage = Temp_Tree()
        self.source_trees = []
        self.backup_count = 0
        self.backup_name = 'backup_test'

        self.get_environ()

    def tearDown(self):

        super(BaseFreezerTest, self).tearDown()

        for tree in self.source_trees:
            tree.cleanup()

        self.storage.cleanup()

    @classmethod
    def get_auth_url(cls):
        return cls.os_primary.auth_provider.auth_client.auth_url[:-len(
            '/auth/tokens')]

    @classmethod
    def setup_clients(cls):
        super(BaseFreezerTest, cls).setup_clients()
        cls.get_environ()

    @classmethod
    def get_environ(cls):
        os.environ['OS_PASSWORD'] = cls.os_primary.credentials.password
        os.environ['OS_USERNAME'] = cls.os_primary.credentials.username
        os.environ['OS_PROJECT_NAME'] = cls.os_primary.credentials.tenant_name
        os.environ['OS_TENANT_NAME'] = cls.os_primary.credentials.tenant_name
        os.environ['OS_PROJECT_DOMAIN_NAME'] = \
            cls.os_primary.credentials.project_domain_name
        os.environ['OS_USER_DOMAIN_NAME'] = \
            cls.os_primary.credentials.user_domain_name

        # Allow developers to set OS_AUTH_URL when developing so that
        # Keystone may be on a host other than localhost.
        if 'OS_AUTH_URL' not in os.environ:
            os.environ['OS_AUTH_URL'] = cls.get_auth_url()

        # Mac OS X uses gtar located in /usr/local/bin
        os.environ['PATH'] = '/usr/local/bin:' + os.environ['PATH']

        return os.environ

    def run_subprocess(self, sub_process_args, fail_message):

        proc = subprocess.Popen(sub_process_args,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                env=self.environ, shell=False)

        out, err = proc.communicate()

        self.assertEqual(0, proc.returncode,
                         fail_message + " Output: {0}. "
                                        "Error: {1}".format(out, err))

        self.assertEqual('', err,
                         fail_message + " Output: {0}. "
                                        "Error: {1}".format(out, err))

    def create_local_backup(self, hostname=None, compression=None,
                            consistency_check=None, incremental=True,
                            always_level=None, restart_always_level=None,
                            max_level=None):
        """Creates a new backup with the given parameters.

        The backup will immediately be created using a randomly-generated
        source tree on the local filesystem, and will be stored in a random
        temporary directory using the 'local' storage mode. All generated data
        files will be automatically removed during `tearDown()`, though
        implementations are responsible for cleaning up any additional copies
        or restores created via other methods.

        :param hostname: if set, set `--hostname` to the given value
        :param compression: if set, set `--compression` to the given value
        :param consistency_check: if True, set `--consistency_check`
        :param incremental: if False, set `--no-incremental`
        :param always_level: sets `--always-level` to the given value
        :param restart_always_level: sets `--restart-always-level`
        :param max_level: sets `--max-level` to the given value
        :return: the path to the stored backup metadata
        """
        metadata_path = os.path.join(
            self.storage.path,
            'metadata-{}.json'.format(self.backup_count))
        self.backup_count += 1

        tree = Temp_Tree()
        tree.add_random_data()
        self.source_trees.append(tree)

        backup_args = [
            'freezer-agent',
            '--path-to-backup', tree.path,
            '--container', self.storage.path,
            '--backup-name', self.backup_name,
            '--storage', 'local',
            '--metadata-out', metadata_path,
        ]

        if hostname:
            backup_args += ['--hostname', hostname]

        if compression:
            backup_args += ['--compression', compression]

        if consistency_check:
            backup_args += ['--consistency-check']

        if incremental:
            if always_level is not None:
                backup_args += ['--always-level', str(always_level)]

            if max_level is not None:
                backup_args += ['--max-level', str(max_level)]

            if restart_always_level:
                backup_args += ['--restart-always-level',
                                str(restart_always_level)]
        else:
            backup_args += ['--no-incremental', 'NO_INCREMENTAL']

        self.run_subprocess(backup_args, 'Test backup to local storage.')

        return metadata_path

    def create_mutated_backup(self, days_old=30, **kwargs):
        """Create a local backup with a mutated timestamp

        This creates a new backup using `create_local_backup()`, modifies it
        using `mutate_timestamp()`, and then returns the resulting (loaded)
        metadata dict.

        :param days_old: the age of the backup to create
        :param kwargs: arguments to pass to `create_local_backup()`
        :return: the loaded metadata
        """
        metadata_path = self.create_local_backup(**kwargs)

        metadata = load_metadata(metadata_path)
        mutate_timestamp(metadata, days_old)
        save_metadata(metadata, metadata_path)

        return metadata



"""
(c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import unittest

import mock
from mock import patch

from freezer.snapshot import lvm


class Test_lvm_snap_remove(unittest.TestCase):
    @patch('freezer.snapshot.lvm.os')
    @patch('freezer.snapshot.lvm._umount')
    @patch('freezer.snapshot.lvm._lvremove')
    def test_return_none_on_success(self, mock_lvremove, mock_umount, mock_os):
        backup_opt = mock.Mock()
        backup_opt.lvm_volgroup = 'one'
        backup_opt.lvm_snapname = 'two'
        self.assertIsNone(lvm.lvm_snap_remove(backup_opt))


class Test_lvm_snap(unittest.TestCase):
    @patch('freezer.snapshot.lvm.get_lvm_info')
    @patch('freezer.snapshot.lvm.utils.create_dir')
    def test_with_snapshot_opt_simple_sets_correct_path_and_raises_on_perm(
            self, mock_create_dir, mock_get_lvm_info):
        mock_get_lvm_info.return_value = {
            'volgroup': 'lvm_volgroup',
            'srcvol': 'lvm_device',
            'snap_path': 'snap_path'}

        backup_opt = mock.Mock()
        backup_opt.snapshot = True
        backup_opt.path_to_backup = '/just/a/path'
        backup_opt.lvm_dirmount = '/var/mountpoint'
        backup_opt.lvm_snapperm = 'invalid_value'

        with self.assertRaises(Exception) as cm:  # noqa
            lvm.lvm_snap(backup_opt)
        the_exception = cm.exception
        self.assertIn('Invalid value for option lvm-snap-perm',
                      str(the_exception))

    @patch('freezer.snapshot.lvm.validate_lvm_params')
    @patch('freezer.snapshot.lvm.subprocess.Popen')
    @patch('freezer.snapshot.lvm.get_vol_fs_type')
    @patch('freezer.snapshot.lvm.get_lvm_info')
    @patch('freezer.snapshot.lvm.utils.create_dir')
    def test_ok(self, mock_create_dir, mock_get_lvm_info, mock_get_vol_fs_type,
                mock_popen, mock_validate_lvm_params):
        mock_get_lvm_info.return_value = {
            'volgroup': 'lvm_volgroup',
            'srcvol': 'lvm_device',
            'snap_path': 'snap_path'}
        mock_process = mock.Mock()
        mock_process.communicate.return_value = '', ''
        mock_process.returncode = 0
        mock_popen.return_value = mock_process

        backup_opt = mock.Mock()
        backup_opt.snapshot = True
        backup_opt.path_to_backup = '/just/a/path'
        backup_opt.lvm_dirmount = '/var/mountpoint'
        backup_opt.lvm_snapperm = 'ro'
        backup_opt.lvm_volgroup = ''
        backup_opt.lvm_srcvol = ''
        backup_opt.lvm_snapsize = '1G'

        self.assertTrue(lvm.lvm_snap(backup_opt))

    @patch('freezer.snapshot.lvm.validate_lvm_params')
    @patch('freezer.snapshot.lvm.subprocess.Popen')
    @patch('freezer.snapshot.lvm.get_vol_fs_type')
    @patch('freezer.snapshot.lvm.get_lvm_info')
    @patch('freezer.utils.utils.create_dir')
    def test_snapshot_fails(self, mock_create_dir, mock_get_lvm_info,
                            mock_get_vol_fs_type, mock_popen,
                            mock_validate_lvm_params):
        mock_get_lvm_info.return_value = {
            'volgroup': 'lvm_volgroup',
            'srcvol': 'lvm_device',
            'snap_path': 'snap_path'}
        mock_process = mock.Mock()
        mock_process.communicate.return_value = '', ''
        mock_process.returncode = 1
        mock_popen.return_value = mock_process

        backup_opt = mock.Mock()
        backup_opt.snapshot = True
        backup_opt.path_to_backup = '/just/a/path'
        backup_opt.lvm_dirmount = '/var/mountpoint'
        backup_opt.lvm_snapperm = 'ro'
        backup_opt.lvm_snapsize = '1G'

        with self.assertRaises(Exception) as cm:  # noqa
            lvm.lvm_snap(backup_opt)
        the_exception = cm.exception
        self.assertIn('lvm snapshot creation error', str(the_exception))

    @patch('freezer.snapshot.lvm.lvm_snap_remove')
    @patch('freezer.snapshot.lvm.validate_lvm_params')
    @patch('freezer.snapshot.lvm.subprocess.Popen')
    @patch('freezer.snapshot.lvm.get_vol_fs_type')
    @patch('freezer.snapshot.lvm.get_lvm_info')
    @patch('freezer.utils.utils.create_dir')
    def test_already_mounted(self, mock_create_dir, mock_get_lvm_info,
                             mock_get_vol_fs_type,
                             mock_popen, mock_validate_lvm_params,
                             lvm_snap_remove):
        mock_get_vol_fs_type.return_value = 'xfs'
        mock_get_lvm_info.return_value = {
            'volgroup': 'lvm_volgroup',
            'srcvol': 'lvm_device',
            'snap_path': 'snap_path'}
        mock_process = mock.Mock()
        mock_process.communicate.return_value = '', 'already mounted'
        mock_process.returncode = 0
        mock_popen.return_value = mock_process

        backup_opt = mock.Mock()
        backup_opt.snapshot = True
        backup_opt.path_to_backup = '/just/a/path'
        backup_opt.lvm_dirmount = '/var/mountpoint'
        backup_opt.lvm_snapperm = 'ro'
        backup_opt.lvm_snapsize = '1G'

        self.assertTrue(lvm.lvm_snap(backup_opt))

    @patch('freezer.snapshot.lvm.lvm_snap_remove')
    @patch('freezer.snapshot.lvm.subprocess.Popen')
    @patch('freezer.snapshot.lvm.get_vol_fs_type')
    @patch('freezer.snapshot.lvm.get_lvm_info')
    @patch('freezer.snapshot.lvm.utils.create_dir')
    def test_snapshot_mount_error_raises_Exception(self,
                                                   mock_create_dir,
                                                   mock_get_lvm_info,
                                                   mock_get_vol_fs_type,
                                                   mock_popen,
                                                   mock_lvm_snap_remove):
        mock_get_vol_fs_type.return_value = 'xfs'
        mock_get_lvm_info.return_value = {
            'volgroup': 'lvm_volgroup',
            'srcvol': 'lvm_device',
            'snap_path': 'snap_path'}
        mock_lvcreate_process, mock_mount_process = mock.Mock(), mock.Mock()

        mock_lvcreate_process.communicate.return_value = '', ''
        mock_lvcreate_process.returncode = 0

        mock_mount_process.communicate.return_value = '', 'mount error'
        mock_mount_process.returncode = 1

        mock_popen.side_effect = [mock_lvcreate_process, mock_mount_process]

        backup_opt = mock.Mock()
        backup_opt.snapshot = True
        backup_opt.path_to_backup = '/just/a/path'
        backup_opt.lvm_dirmount = '/var/mountpoint'
        backup_opt.lvm_snapperm = 'ro'
        backup_opt.lvm_snapsize = '1G'

        with self.assertRaises(Exception) as cm:  # noqa
            lvm.lvm_snap(backup_opt)
        the_exception = cm.exception
        self.assertIn('lvm snapshot mounting error', str(the_exception))

        mock_lvm_snap_remove.assert_called_once_with(backup_opt)


# class Test_get_lvm_info(unittest.TestCase):
#     @patch('freezer.snapshot.lvm.lvm_guess')
#     @patch('freezer.snapshot.lvm.utils.get_mount_from_path')
#     def test_using_guess(self, mock_get_mount_from_path, mock_lvm_guess):
#         mock_get_mount_from_path.return_value = ('/home/somedir',
#                                                  'some-snap-path')
#         mock_lvm_guess.return_value = 'vg_test', 'lv_test', 'lvm_device'
#         mounts = (
#             '/dev/mapper/vg_prova-lv_prova_vol1 /home/pippo ext4 rw,'
#             'relatime,data=ordered 0 0'
#         )
#         mocked_open_function = mock.mock_open(read_data=mounts)
#
#         with patch("__builtin__.open", mocked_open_function):
#             res = lvm.get_lvm_info('lvm_auto_snap_value')
#
#         expected_result = {'volgroup': 'vg_test',
#                            'snap_path': 'some-snap-path',
#                            'srcvol': 'lvm_device'}
#         self.assertEqual(res, expected_result)
#
#     @patch('freezer.snapshot.lvm.subprocess.Popen')
#     @patch('freezer.snapshot.lvm.lvm_guess')
#     @patch('freezer.snapshot.lvm.utils.get_mount_from_path')
#     def test_using_mount(self, mock_get_mount_from_path, mock_lvm_guess,
#                          mock_popen):
#         mock_get_mount_from_path.return_value = ('/home/somedir',
#                                                  'some-snap-path')
#         mock_lvm_guess.side_effect = [(None, None, None),
#                                       ('vg_test', 'lv_test', 'lvm_device')]
#         mounts = (
#             '/dev/mapper/vg_prova-lv_prova_vol1 /home/pippo ext4 rw,'
#             'relatime,data=ordered 0 0'
#         )
#         mocked_open_function = mock.mock_open(read_data=mounts)
#         mock_process = mock.Mock()
#         mock_process.returncode = 0
#         mock_popen.return_value = mock_process
#         mock_process.communicate.return_value = '', ''
#
#         with patch("__builtin__.open", mocked_open_function):
#             res = lvm.get_lvm_info('lvm_auto_snap_value')
#
#         expected_result = {'volgroup': 'vg_test',
#                            'snap_path': 'some-snap-path',
#                            'srcvol': 'lvm_device'}
#         self.assertEqual(res, expected_result)
#
#     @patch('freezer.snapshot.lvm.subprocess.Popen')
#     @patch('freezer.snapshot.lvm.lvm_guess')
#     @patch('freezer.snapshot.lvm.utils.get_mount_from_path')
#     def test_raises_Exception_when_info_not_found(
#             self, mock_get_mount_from_path, mock_lvm_guess, mock_popen):
#         mock_get_mount_from_path.return_value = ('/home/somedir',
#                                                  'some-snap-path')
#         mock_lvm_guess.return_value = None, None, None
#         mounts = (
#             '/dev/mapper/vg_prova-lv_prova_vol1 /home/pippo ext4 rw,'
#             'relatime,data=ordered 0 0'
#         )
#         mocked_open_function = mock.mock_open(read_data=mounts)
#         mock_process = mock.Mock()
#         mock_lvm_guess.return_value = None, None, None
#         mock_process.communicate.return_value = '', ''
#         mock_popen.return_value = mock_process
#
#         with patch("__builtin__.open", mocked_open_function):
#             self.assertRaises(Exception, lvm.get_lvm_info,
#                               'lvm_auto_snap_value')


class Test_lvm_guess(unittest.TestCase):
    def test_no_match(self):
        mount_points = []
        mount_point_path = '/home/pippo'
        source = '/proc/mounts'

        res = lvm.lvm_guess(mount_point_path, mount_points, source)

        expected_result = (None, None, None)
        self.assertEqual(res, expected_result)

    def test_unsing_proc_mounts(self):
        mount_points = [
            'rootfs / rootfs rw 0 0\n',
            'sysfs /sys sysfs rw,nosuid,nodev,noexec,relatime 0 0\n',
            'proc /proc proc rw,nosuid,nodev,noexec,relatime 0 0\n',
            'udev /dev devtmpfs rw,relatime,size=2010616k,nr_inodes=502654,'
            'mode=755 0 0\n',
            'devpts /dev/pts devpts rw,nosuid,noexec,relatime,gid=5,mode=620,'
            'ptmxmode=000 0 0\n',
            'tmpfs /run tmpfs rw,nosuid,noexec,relatime,size=404836k,mode=755 '
            '0 0\n',
            '/dev/mapper/fabuntu--vg-root / ext4 rw,relatime,'
            'errors=remount-ro,data=ordered 0 0\n',
            'none /sys/fs/cgroup tmpfs rw,relatime,size=4k,mode=755 0 0\n',
            'none /sys/fs/fuse/connections fusectl rw,relatime 0 0\n',
            'none /sys/kernel/debug debugfs rw,relatime 0 0\n',
            'none /sys/kernel/security securityfs rw,relatime 0 0\n',
            'cgroup /sys/fs/cgroup/cpuset cgroup rw,relatime,cpuset 0 0\n',
            'cgroup /sys/fs/cgroup/cpu cgroup rw,relatime,cpu 0 0\n',
            'cgroup /sys/fs/cgroup/cpuacct cgroup rw,relatime,cpuacct 0 0\n',
            'cgroup /sys/fs/cgroup/memory cgroup rw,relatime,memory 0 0\n',
            'none /run/lock tmpfs rw,nosuid,nodev,noexec,relatime,size=5120k '
            '0 0\n',
            'none /run/shm tmpfs rw,nosuid,nodev,relatime 0 0\n',
            'none /run/user tmpfs rw,nosuid,nodev,noexec,relatime,'
            'size=102400k,mode=755 0 0\n',
            'cgroup /sys/fs/cgroup/devices cgroup rw,relatime,devices 0 0\n',
            'none /sys/fs/pstore pstore rw,relatime 0 0\n',
            'cgroup /sys/fs/cgroup/freezer cgroup rw,relatime,freezer 0 0\n',
            'cgroup /sys/fs/cgroup/blkio cgroup rw,relatime,blkio 0 0\n',
            'cgroup /sys/fs/cgroup/perf_event cgroup rw,relatime,perf_event '
            '0 0\n',
            'cgroup /sys/fs/cgroup/hugetlb cgroup rw,relatime,hugetlb 0 0\n',
            '/dev/sda1 /boot ext2 rw,relatime 0 0\n',
            'systemd /sys/fs/cgroup/systemd cgroup rw,nosuid,nodev,noexec,'
            'relatime,name=systemd 0 0\n',
            '/dev/mapper/vg_prova-lv_prova_vol1 /home/pippo ext4 rw,relatime,'
            'data=ordered 0 0\n'
        ]
        mount_point_path = '/home/pippo'
        source = '/proc/mounts'

        res = lvm.lvm_guess(mount_point_path, mount_points, source)

        expected_result = (
            'vg_prova', 'lv_prova_vol1', '/dev/vg_prova/lv_prova_vol1')
        self.assertEqual(res, expected_result)

    def test_unsing_mount(self):
        mount_points = [
            '/dev/mapper/fabuntu--vg-root on / type ext4 (rw,errors='
            'remount-ro)',
            'proc on /proc type proc (rw,noexec,nosuid,nodev)',
            'sysfs on /sys type sysfs (rw,noexec,nosuid,nodev)',
            'none on /sys/fs/cgroup type tmpfs (rw)',
            'none on /sys/fs/fuse/connections type fusectl (rw)',
            'none on /sys/kernel/debug type debugfs (rw)',
            'none on /sys/kernel/security type securityfs (rw)',
            'udev on /dev type devtmpfs (rw,mode=0755)',
            'devpts on /dev/pts type devpts (rw,noexec,nosuid,gid=5,'
            'mode=0620)',
            'tmpfs on /run type tmpfs (rw,noexec,nosuid,size=10%,mode=0755)',
            'none on /run/lock type tmpfs (rw,noexec,nosuid,nodev,'
            'size=5242880)',
            'none on /run/shm type tmpfs (rw,nosuid,nodev)',
            'none on /run/user type tmpfs (rw,noexec,nosuid,nodev,'
            'size=104857600,mode=0755)',
            'none on /sys/fs/pstore type pstore (rw)',
            'cgroup on /sys/fs/cgroup/cpuset type cgroup (rw,relatime,cpuset)',
            'cgroup on /sys/fs/cgroup/cpu type cgroup (rw,relatime,cpu)',
            'cgroup on /sys/fs/cgroup/cpuacct type cgroup (rw,relatime,'
            'cpuacct)',
            'cgroup on /sys/fs/cgroup/memory type cgroup (rw,relatime,memory)',
            'cgroup on /sys/fs/cgroup/devices type cgroup (rw,relatime,'
            'devices)',
            'cgroup on /sys/fs/cgroup/freezer type cgroup (rw,relatime,'
            'freezer)',
            'cgroup on /sys/fs/cgroup/blkio type cgroup (rw,relatime,blkio)',
            'cgroup on /sys/fs/cgroup/perf_event type cgroup (rw,relatime,'
            'perf_event)',
            'cgroup on /sys/fs/cgroup/hugetlb type cgroup (rw,relatime,'
            'hugetlb)',
            '/dev/sda1 on /boot type ext2 (rw)',
            'systemd on /sys/fs/cgroup/systemd type cgroup (rw,noexec,nosuid,'
            'nodev,none,name=systemd)',
            '/dev/mapper/vg_prova-lv_prova_vol1 on /home/pippo type ext4 (rw)',
            '']
        mount_point_path = '/home/pippo'
        source = 'mount'

        res = lvm.lvm_guess(mount_point_path, mount_points, source)

        expected_result = ('vg_prova', 'lv_prova_vol1',
                           '/dev/vg_prova/lv_prova_vol1')
        self.assertEqual(res, expected_result)


class Test_validate_lvm_params(unittest.TestCase):
    def setUp(self):
        self.backup_opt = mock.Mock()
        self.backup_opt.lvm_snapperm = 'ro'
        self.backup_opt.path_to_backup = '/path/to/backup'
        self.backup_opt.lvm_srcvol = '/lvm/srcvol'
        self.backup_opt.lvm_volgroup = 'vg_burger'

    def test_return_true_on_lvm_configuration_valid(self):
        self.backup_opt.lvm_srcvol = ''
        self.backup_opt.lvm_volgroup = ''
        self.assertFalse(lvm.validate_lvm_params(self.backup_opt))

    def test_return_false_on_lvm_not_required(self):
        self.assertTrue(lvm.validate_lvm_params(self.backup_opt))

    def test_raises_Exception_on_snapperm_invalid(self):
        self.backup_opt.lvm_snapperm = 'squeezeme'
        self.assertRaises(Exception, lvm.validate_lvm_params,
                          self.backup_opt)  # noqa

    def test_raises_Exception_on_no_pathtobackup(self):
        self.backup_opt.path_to_backup = ''
        self.assertRaises(Exception, lvm.validate_lvm_params,
                          self.backup_opt)  # noqa

    def test_raises_Exception_on_no_lvmsrcvol(self):
        self.backup_opt.lvm_srcvol = ''
        self.assertRaises(Exception, lvm.validate_lvm_params,
                          self.backup_opt)  # noqa

    def test_raises_Exception_on_no_lvmvolgrp(self):
        self.backup_opt.lvm_volgroup = ''
        self.assertRaises(Exception, lvm.validate_lvm_params,
                          self.backup_opt)  # noqa


class Test_umount(unittest.TestCase):
    @patch('freezer.snapshot.lvm.subprocess.Popen')
    @patch('freezer.snapshot.lvm.os')
    def test_return_none_on_success(self, mock_os, mock_popen):
        mock_process = mock.Mock()
        mock_process.communicate.return_value = '', ''
        mock_process.returncode = 0
        mock_popen.return_value = mock_process
        mock_os.rmdir.return_value = None
        self.assertIsNone(lvm._umount('path'))

    @patch('freezer.snapshot.lvm.subprocess.Popen')
    @patch('freezer.snapshot.lvm.os')
    def test_raises_on_popen_returncode_not_0(self, mock_os, mock_popen):
        mock_process = mock.Mock()
        mock_process.communicate.return_value = '', ''
        mock_process.returncode = 1
        mock_popen.return_value = mock_process
        mock_os.rmdir.return_value = None
        self.assertRaises(Exception, lvm._umount, 'path')  # noqa


class Test_lvremove(unittest.TestCase):
    @patch('freezer.snapshot.lvm.subprocess.Popen')
    def test_return_none_on_success(self, mock_popen):
        mock_process = mock.Mock()
        mock_process.communicate.return_value = '', ''
        mock_process.returncode = 0
        mock_popen.return_value = mock_process
        self.assertIsNone(lvm._lvremove('logicalvolume'))

    @patch('freezer.snapshot.lvm.subprocess.Popen')
    def test_raises_on_popen_returncode_not_0(self, mock_popen):
        mock_process = mock.Mock()
        mock_process.communicate.return_value = '', ''
        mock_process.returncode = 1
        mock_popen.return_value = mock_process
        self.assertRaises(Exception, lvm._lvremove, 'path')  # noqa



"""
(c) Copyright 2016 Hewlett-Packard Development Enterprise, L.P.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""
from concurrent import futures
import os

from oslo_config import cfg
from oslo_log import log

from freezer.common import client_manager
from freezer.engine import engine
from freezer.utils import utils

import json

LOG = log.getLogger(__name__)
CONF = cfg.CONF


class NovaEngine(engine.BackupEngine):

    def __init__(self, storage, **kwargs):
        super(NovaEngine, self).__init__(storage=storage)
        self.client = client_manager.get_client_manager(CONF)
        self.nova = self.client.create_nova()
        self.glance = self.client.create_glance()
        self.cinder = self.client.create_cinder()
        self.neutron = self.client.create_neutron()
        self.server_info = None

    @property
    def name(self):
        return "nova"

    def stream_image(self, pipe):
        """Reading bytes from a pipe and converting it to a stream-like"""
        try:
            while True:
                chunk = pipe.recv_bytes()
                yield chunk
        except EOFError:
            pass

    def get_nova_tenant(self, project_id):
        # Load info about tenant instances
        if self.storage._type == 'swift':
            swift_connection = self.client.create_swift()
            headers, data = swift_connection.get_object(
                self.storage.storage_path,
                "project_" + project_id)
        elif self.storage._type == 's3':
            bucket_name, object_name = self.get_storage_info(project_id)
            data = self.storage.get_object(
                bucket_name=bucket_name,
                key=object_name
            )['Body'].read()
        elif self.storage._type in ['local', 'ssh']:
            backup_basepath = os.path.join(self.storage.storage_path,
                                           'project_' + project_id)
            with self.storage.open(backup_basepath, 'rb') as backup_file:
                data = backup_file.readline()

        return json.loads(data)

    def restore_nova_tenant(self, project_id, hostname_backup_name,
                            overwrite, recent_to_date):

        instance_ids = self.get_nova_tenant(project_id)
        for instance_id in instance_ids:
            LOG.info("Restore nova instance ID: {0} from container {1}".
                     format(instance_id, self.storage.storage_path))
            backup_name = os.path.join(hostname_backup_name,
                                       instance_id)
            self.restore(
                hostname_backup_name=backup_name,
                restore_resource=instance_id,
                overwrite=overwrite,
                recent_to_date=recent_to_date)

    def restore_level(self, restore_resource, read_pipe, backup, except_queue):
        try:
            metadata = backup.metadata()
            engine_metadata = backup.engine_metadata()
            server_info = metadata.get('server', {})
            length = int(engine_metadata.get('length'))
            available_networks = server_info.get('addresses')
            nova_networks = self.neutron.list_networks()['networks']

            net_names = [network for network, _ in
                         available_networks.iteritems()]
            match_networks = [{"net-id": network.get('id')} for network in
                              nova_networks
                              if network.get('name') in net_names]

            stream = self.stream_image(read_pipe)
            data = utils.ReSizeStream(stream, length, 1)
            image = self.client.create_image(
                "Restore: {0}".format(
                    server_info.get('name', server_info.get('id', None))
                ),
                'bare',
                'raw',
                data=data
            )

            utils.wait_for(
                NovaEngine.image_active,
                1,
                CONF.timeout,
                message="Waiting for image to finish uploading {0} and become"
                        " active".format(image.id),
                kwargs={"glance_client": self.glance, "image_id": image.id}
            )
            server = self.nova.servers.create(
                name=server_info.get('name'),
                flavor=server_info['flavor']['id'],
                image=image.id,
                nics=match_networks
            )
            return server
        except Exception as e:
            LOG.exception(e)
            except_queue.put(e)
            raise

    def backup_nova_tenant(self, project_id, hostname_backup_name,
                           no_incremental, max_level, always_level,
                           restart_always_level):
        instance_ids = [server.id for server in
                        self.nova.servers.list(detailed=False)]
        data = json.dumps(instance_ids)
        LOG.info("Saving information about instances {0}".format(data))

        if self.storage._type == 'swift':
            swift_connection = self.client.create_swift()
            swift_connection.put_object(self.storage.storage_path,
                                        "project_{0}".format(project_id),
                                        data)
        elif self.storage._type == 's3':
            bucket_name, object_name = self.get_storage_info(project_id)
            self.storage.put_object(
                bucket_name=bucket_name,
                key=object_name,
                data=data
            )
        elif self.storage._type in ['local', 'ssh']:
            backup_basepath = os.path.join(self.storage.storage_path,
                                           "project_" + project_id)
            with self.storage.open(backup_basepath, 'wb') as backup_file:
                backup_file.write(data)

        executor = futures.ThreadPoolExecutor(
            max_workers=len(instance_ids))
        futures_list = []
        for instance_id in instance_ids:
            LOG.info("Backup nova instance ID: {0} to container {1}".
                     format(instance_id, self.storage.storage_path))
            backup_name = os.path.join(hostname_backup_name,
                                       instance_id)

            futures_list.append(executor.submit(
                self.backup,
                backup_resource=instance_id,
                hostname_backup_name=backup_name,
                no_incremental=no_incremental,
                max_level=max_level,
                always_level=always_level,
                restart_always_level=restart_always_level))

        futures.wait(futures_list, CONF.timeout)

    def get_storage_info(self, project_id):
        if self.storage.get_object_prefix() != '':
            object_name = "{0}/project_{1}".format(
                self.storage.get_object_prefix(),
                project_id
            )
        else:
            object_name = "project_{0}".format(project_id)
        return self.storage.get_bucket_name(), object_name

    def backup_data(self, backup_resource, manifest_path):
        server = self.nova.servers.get(backup_resource)
        if not server:
            raise Exception("Server not found {0}".format(backup_resource))

        def instance_finish_task():
            server = self.nova.servers.get(backup_resource)
            return not server.__dict__['OS-EXT-STS:task_state']

        utils.wait_for(
            instance_finish_task, 1, CONF.timeout,
            message="Waiting for instance {0} to finish {1} to start the "
                    "snapshot process".format(
                        backup_resource,
                        server.__dict__['OS-EXT-STS:task_state']
                    )
        )
        image_id = self.nova.servers.create_image(
            server,
            "snapshot_of_{0}".format(backup_resource)
        )
        image = self.glance.images.get(image_id)
        if not image:
            raise Exception(
                "Image {0} is not created or can't be found.".format(image_id)
            )
        # wait a bit for the snapshot to be taken and completely uploaded
        # to glance.
        utils.wait_for(
            NovaEngine.image_active,
            1,
            100,
            message="Waiting for instnace {0} snapshot to become "
                    "active".format(backup_resource),
            kwargs={"glance_client": self.glance, "image_id": image_id}
        )

        image = self.glance.images.get(image_id)
        image_temporary_snapshot_id = None
        copied_volume = None
        image_info = getattr(server, "image", None)
        if image_info is not None and isinstance(image_info, dict):
            LOG.info('Image type instance backup')
            boot_device_type = "image"
            stream = self.client.download_image(image)
        else:
            LOG.info('Volume or snapshot type instance backup')
            boot_device_type = "volume"
            image_block_mapping_info = image.get("block_device_mapping")
            image_block_mapping = json.loads(image_block_mapping_info) \
                if image_block_mapping_info else None
            image_temporary_snapshot_id = \
                image_block_mapping[0]['snapshot_id'] \
                if image_block_mapping else None
            copied_volume = self.client.do_copy_volume(
                self.cinder.volume_snapshots.get(
                    image_temporary_snapshot_id))
            LOG.debug("Deleting temporary glance image "
                      "generated by snapshot")
            self.glance.images.delete(image.id)
            LOG.debug("Creation temporary glance image")
            image = self.client.make_glance_image(
                copied_volume.id, copied_volume)
            LOG.debug("Download temporary glance image {0}".format(image.id))
            stream = self.client.download_image(image)

        LOG.info("Uploading image to storage path")
        headers = {"server_name": server.name,
                   "flavour_id": str(server.flavor.get('id')),
                   'length': str(len(stream)),
                   "boot_device_type": boot_device_type}
        self.set_tenant_meta(manifest_path, headers)
        for chunk in stream:
            yield chunk

        if image_temporary_snapshot_id is not None:
            LOG.info("Deleting temporary snapshot {0}"
                     .format(image_temporary_snapshot_id))
            self.cinder.volume_snapshots.delete(image_temporary_snapshot_id)
        if copied_volume is not None:
            LOG.info("Deleting temporary copied volume {0}"
                     .format(copied_volume.id))
            self.cinder.volumes.delete(copied_volume)

        LOG.info("Deleting temporary image {0}".format(image.id))
        self.glance.images.delete(image.id)

    @staticmethod
    def image_active(glance_client, image_id):
        """Check if the image is in the active state or not"""
        image = glance_client.images.get(image_id)
        return image.status == 'active'

    def metadata(self, backup_resource):
        """Construct metadata"""
        server_info = self.nova.servers.get(backup_resource).to_dict()

        return {
            "engine_name": self.name,
            "server": server_info,
        }

    def set_tenant_meta(self, path, metadata):
        """push data to the manifest file"""
        with open(path, 'wb') as fb:
            fb.writelines(json.dumps(metadata))

    def get_tenant_meta(self, path):
        with open(path, 'rb') as fb:
            json.loads(fb.read())



"""
(c) Copyright 2016 Hewlett-Packard Development Enterprise, L.P.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""


class EngineException(Exception):
    msg = "An unknown error occurred."

    def __init__(self, message=None, **kwargs):
        if not message:
            message = self.msg
        super(EngineException, self).__init__(message, kwargs)


class EngineNotFound(EngineException):

    def __init__(self, message):
        super(EngineNotFound, self).__init__(message)



"""
(c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Freezer Tar related functions
"""
from freezer.utils import utils


class TarCommandBuilder(object):
    """
    Building a tar cmd command. To build command invoke method build.
    """

    UNIX_TEMPLATE = (
        "{gnutar_path} --create {algo} --warning=none --no-check-device "
        "--one-file-system --preserve-permissions --same-owner "
        "--seek --ignore-failed-read")

    # local windows template
    WINDOWS_TEMPLATE = (
        "{gnutar_path} -c {algo} --incremental --unlink-first "
        "--ignore-zeros")

    LISTED_TEMPLATE = "{tar_command} --listed-incremental={listed_incremental}"

    DEREFERENCE_MODE = {'soft': '--dereference',
                        'hard': '--hard-dereference',
                        'all': '--hard-dereference --dereference'}

    def __init__(self, filepath, compression_algo, is_windows, tar_path=None):
        self.tar_path = tar_path or utils.tar_path()
        self.dereference = ''
        self.listed_incremental = None
        self.exclude = ''
        self.openssl_path = None
        self.encrypt_pass_file = None
        self.output_file = None
        self.filepath = filepath
        self.compression_algo = get_tar_flag_from_algo(compression_algo)
        self.is_windows = is_windows

    def set_listed_incremental(self, absolute_path):
        self.listed_incremental = absolute_path

    def set_exclude(self, exclude):
        self.exclude = exclude

    def set_dereference(self, mode):
        """
        Dereference hard and soft links according option choices.
            'soft' dereference soft links,
            'hard' dereference hardlinks,
            'all' dereference both.
            Default 'None'.
        """
        self.dereference = self.DEREFERENCE_MODE[mode]

    def set_encryption(self, encrypt_pass_file, openssl_path=None):
        self.openssl_path = openssl_path or utils.openssl_path()
        self.encrypt_pass_file = encrypt_pass_file

    def build(self):
        if self.is_windows:
            tar_command = self.WINDOWS_TEMPLATE.format(
                gnutar_path=self.tar_path, algo=self.compression_algo)
        else:
            tar_command = self.UNIX_TEMPLATE.format(
                gnutar_path=self.tar_path, algo=self.compression_algo)

        if self.dereference:
            tar_command = "{0} {1}".format(tar_command, self.dereference)

        if self.listed_incremental:
            tar_command = self.LISTED_TEMPLATE.format(
                tar_command=tar_command,
                listed_incremental=self.listed_incremental)

        if self.exclude:
            tar_command = '{tar_command} --exclude="{exclude}"'.format(
                tar_command=tar_command, exclude=self.exclude)

        tar_command = '{0} {1}'.format(tar_command, self.filepath)

        if self.encrypt_pass_file:
            openssl_cmd = "{openssl_path} enc -aes-256-cfb -pass file:{file}"\
                .format(openssl_path=self.openssl_path,
                        file=self.encrypt_pass_file)
            tar_command = '{0} | {1} && exit ${{PIPESTATUS[0]}}'\
                .format(tar_command, openssl_cmd)

        return tar_command


class TarCommandRestoreBuilder(object):
    WINDOWS_TEMPLATE = '{0} -x {1} --incremental --unlink-first ' \
        '--ignore-zeros'
    DRY_RUN_TEMPLATE = '{0} {1} --incremental --list ' \
        '--ignore-zeros --warning=none'
    UNIX_TEMPLATE = '{0} {1} --incremental --extract ' \
        '--ignore-zeros --warning=none --overwrite --directory {2}'
    OPENSSL_DEC = "{openssl_path} enc -d -aes-256-cfb -pass file:{file}"

    def __init__(self, restore_path, compression_algo, is_windows,
                 tar_path=None):
        self.dry_run = False
        self.is_windows = False
        self.openssl_path = None
        self.encrypt_pass_file = None
        self.tar_path = tar_path or utils.tar_path()
        self.restore_path = restore_path
        self.compression_algo = get_tar_flag_from_algo(compression_algo)
        self.is_windows = is_windows

    def set_dry_run(self):
        self.dry_run = True

    def set_encryption(self, encrypt_pass_file, openssl_path=None):
        self.openssl_path = openssl_path or utils.openssl_path()
        self.encrypt_pass_file = encrypt_pass_file

    def build(self):
        if self.dry_run:
            tar_command = self.DRY_RUN_TEMPLATE.format(self.tar_path,
                                                       self.compression_algo)
        elif self.is_windows:
            tar_command = self.WINDOWS_TEMPLATE.format(self.tar_path,
                                                       self.compression_algo)
        else:
            tar_command = self.UNIX_TEMPLATE.format(self.tar_path,
                                                    self.compression_algo,
                                                    self.restore_path)

        # Check if encryption file is provided and set the openssl decrypt
        # command accordingly
        if self.encrypt_pass_file:
            openssl_cmd = self.OPENSSL_DEC.format(
                openssl_path=self.openssl_path,
                file=self.encrypt_pass_file)
            tar_command = '{0} | {1} && exit ${{PIPESTATUS[0]}}'\
                .format(openssl_cmd, tar_command)
        return tar_command


def get_tar_flag_from_algo(compression):
    algo = {
        'gzip': '-z',
        'bzip2': '-j',
        'xz': '-J',
    }
    compression_exec = utils.get_executable_path(compression)
    if not compression_exec:
        raise Exception("Critical Error: {0} executable not found ".
                        format(compression))
    return algo.get(compression)



"""Freezer rsync incremental engine

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import getpass
import grp
import json
import os
import pwd
import Queue
import re
import stat
import sys
import threading

from oslo_log import log
from six.moves import cStringIO

from freezer.engine import engine
from freezer.engine.rsync import pyrsync
from freezer.utils import compress
from freezer.utils import crypt
from freezer.utils import winutils

LOG = log.getLogger(__name__)

# Data block size of used by rsync to generate signature
RSYNC_BLOCK_SIZE = 4096
# Version of the meta data structure format
RSYNC_DATA_STRUCT_VERSION = 1
# Rsync main block size for streams, 32MB (1024*1024*32)
RSYNC_BLOCK_BUFF_SIZE = 33554432
# Files type where file data content can be backed up or restored
REG_FILE = ('r', 'u')


class RsyncEngine(engine.BackupEngine):

    def __init__(
            self, compression, symlinks, exclude, storage,
            max_segment_size, encrypt_key=None,
            dry_run=False):
        self.compression_algo = compression
        self.encrypt_pass_file = encrypt_key
        self.dereference_symlink = symlinks
        self.exclude = exclude
        self.storage = storage
        self.is_windows = winutils.is_windows()
        self.dry_run = dry_run
        self.max_segment_size = max_segment_size
        # Compression and encryption objects
        self.compressor = None
        self.cipher = None
        super(RsyncEngine, self).__init__(storage=storage)

    @property
    def name(self):
        return "rsync"

    def metadata(self, *args):
        return {
            "engine_name": self.name,
            "compression": self.compression_algo,
            # the encrypt_pass_file might be key content so we need to convert
            # to boolean
            "encryption": bool(self.encrypt_pass_file)
        }

    def backup_data(self, backup_resource, manifest_path):
        """Execute backup using rsync algorithm.

        If an existing rsync meta data is available the backup
        will be incremental, otherwise will be executed a level 0 backup

        :param backup_resource:
        :param manifest_path:
        :return:
        """

        LOG.info("Starting RSYNC engine backup data stream")

        file_read_limit = 0
        data_chunk = b''
        LOG.info(
            'Recursively archiving and compressing files from {}'.format(
                os.getcwd()))

        self.compressor = compress.Compressor(self.compression_algo)

        if self.encrypt_pass_file:
            self.cipher = crypt.AESEncrypt(self.encrypt_pass_file)
            data_chunk += self.cipher.generate_header()

        rsync_queue = Queue.Queue(maxsize=2)

        t_get_sign_delta = threading.Thread(
            target=self.get_sign_delta,
            args=(
                backup_resource, manifest_path, rsync_queue))
        t_get_sign_delta.daemon = True

        t_get_sign_delta.start()

        while True:
            file_block = rsync_queue.get()

            if file_block is False:
                break
            if len(file_block) == 0:
                continue

            data_chunk += file_block
            file_read_limit += len(file_block)
            if file_read_limit >= self.max_segment_size:
                yield data_chunk
                data_chunk = b''
                file_read_limit = 0

        # Upload segments smaller then max_segment_size
        if len(data_chunk) < self.max_segment_size:
            yield data_chunk

        # Rejoining thread
        t_get_sign_delta.join()

    def restore_level(self, restore_resource, read_pipe, backup, except_queue):
        """Restore the provided file into restore_abs_path.

        Decrypt the file if backup_opt_dict.encrypt_pass_file key is provided.
        Freezer rsync header data structure:

            header_len, RSYNC_DATA_STRUCT_VERSION, file_mode,
            os_stat.st_uid, os_stat.st_gid, os_stat.st_size,
            mtime, ctime, uname, gname, file_type, linkname

        :param restore_resource:
        :param read_pipe:
        :param backup:
        :param except_queue:
        :return:
        """
        try:
            metadata = backup.metadata()
            if (not self.encrypt_pass_file and
                    metadata.get("encryption", False)):
                raise Exception("Cannot restore encrypted backup without key")

            self.compression_algo = metadata.get('compression',
                                                 self.compression_algo)

            if not os.path.exists(restore_resource):
                raise ValueError(
                    'Provided restore path does not exist: {0}'.format(
                        restore_resource))

            if self.dry_run:
                restore_resource = '/dev/null'

            raw_data_chunk = read_pipe.recv_bytes()

            self.compressor = compress.Decompressor(self.compression_algo)

            if self.encrypt_pass_file:
                self.cipher = crypt.AESDecrypt(self.encrypt_pass_file,
                                               raw_data_chunk[:16])
                raw_data_chunk = raw_data_chunk[16:]

            data_chunk = self.process_restore_data(raw_data_chunk)

            header_str = r'^(\d{1,})\00'
            flushed = False
            while True:
                header_match = re.search(header_str, data_chunk)
                if not header_match and not flushed:
                    try:
                        data_chunk += self.process_restore_data(
                            read_pipe.recv_bytes())
                        continue
                    except EOFError:
                        LOG.info("EOFError: Pipe closed. Flushing buffer...")
                        data_chunk += self.compressor.flush()
                        flushed = True

                if data_chunk and header_match:
                    header_len = int(header_match.group(1))
                    if header_len > len(data_chunk) and not flushed:
                        try:
                            data_chunk += self.process_restore_data(
                                read_pipe.recv_bytes())
                        except EOFError:
                            LOG.info("[*] End of File: Pipe closed. "
                                     "Flushing the buffer.")
                            data_chunk += self.compressor.flush()
                            flushed = True

                    header = data_chunk[:header_len]
                    header_list = header.split('\00')
                    data_chunk = data_chunk[header_len:]
                    data_chunk = self.make_files(
                        header_list, restore_resource, read_pipe,
                        data_chunk, flushed, backup.level)
                else:
                    LOG.info('No more data available...')
                    break
        except Exception as e:
            LOG.exception(e)
            except_queue.put(e)
            raise

    def process_backup_data(self, data, do_compress=True):
        """Compresses and encrypts provided data according to args"""

        if do_compress:
            data = self.compressor.compress(data)

        if self.encrypt_pass_file:
            data = self.cipher.encrypt(data)
        return data

    def process_restore_data(self, data):
        """Decrypts and decompresses provided data according to args"""

        if self.encrypt_pass_file:
            data = self.cipher.decrypt(data)

        data = self.compressor.decompress(data)
        return data

    @staticmethod
    def rsync_gen_delta(file_path_fd, old_file_meta):
        """Get rsync delta for file descriptor provided as arg.

        :param file_path_fd:
        :param old_file_meta:
        :return:
        """

        if not old_file_meta:
            raise StopIteration

        # If the ctime or mtime has changed, the delta is computed
        # data block is returned

        len_deltas = 0
        old_signature = old_file_meta['signature']
        # Get changed blocks index only
        all_changed_indexes = cStringIO()
        file_path_fd.seek(0)
        previous_index = -1
        modified_blocks = []
        for block_index in pyrsync.rsyncdelta(
                file_path_fd,
                (old_signature[0], old_signature[1]),
                RSYNC_BLOCK_SIZE):

            previous_index += 1

            if not isinstance(block_index, int):
                len_deltas += len(block_index)
                all_changed_indexes.write(b'\00{}'.format(previous_index))
                modified_blocks.append(previous_index)

        # Yield the total length data changed blocks

        yield b'\00' + str(len_deltas)
        previous_index_str = all_changed_indexes.getvalue() + b'\00'
        len_previous_index_str = len(previous_index_str) + 1
        # Yield the length of the string that contain all the indexes

        yield b'\00' + str(len_previous_index_str) + b'\00'
        # Yield string containing all the indexes separated by \00

        yield previous_index_str

        # Get blocks of changed data
        file_path_fd.seek(0)
        for block_index in modified_blocks:
            offset = block_index * RSYNC_BLOCK_SIZE
            file_path_fd.seek(offset)
            data_block = file_path_fd.read(RSYNC_BLOCK_SIZE)

            yield data_block

    @staticmethod
    def is_file_modified(old_file_meta, file_meta):
        """Check for changes on inode or file data

        :param old_file_meta: meta data of the previous backup execution
        :param file_meta: meta data of the current backup execution
        :return: True if the file changed, False otherwise
        """

        # Get mtime and ctime from previous backup execution
        old_file_mtime = old_file_meta['inode']['mtime']
        old_file_ctime = old_file_meta['inode']['ctime']
        fs_index_file_mtime = file_meta['inode']['mtime']
        fs_index_file_ctime = file_meta['inode']['ctime']

        # Check if new ctime/mtime is different from the current one
        file_change_flag = None
        if old_file_mtime != fs_index_file_mtime:
            file_change_flag = True
        elif old_file_ctime != fs_index_file_ctime:
            file_change_flag = True

        # TODO(raliev): There is also need to add check size of files
        return file_change_flag

    def write_file(self, file_fd, size, data_chunk, read_pipe, flushed):
        while size:
            written_data = min(len(data_chunk), size)
            file_fd.write(data_chunk[:written_data])
            data_chunk = data_chunk[written_data:]
            size -= written_data
            if size > 0 and not flushed:
                try:
                    data_chunk += self.process_restore_data(
                        read_pipe.recv_bytes())
                except EOFError:
                    LOG.info(
                        "[*] EOF from pipe. Flushing buffer.")
                    data_chunk += self.compressor.flush()
                    flushed = True
                    continue
            elif flushed:
                break
        return data_chunk

    def write_changes_in_file(self, fd_curr_file, data_chunk, read_pipe):
        # Searching for:
        # - the block incremental header string
        # - len of all the changed blocks
        # - len of the index string
        offset_match = re.search(r'^(\00(\d+?)\00(\d+?)\00)', data_chunk)

        if offset_match:
            offset_size = len(offset_match.group(1))
            len_deltas = int(offset_match.group(2))
            len_offset_str = int(offset_match.group(3)) - 1

            no_of_blocks, reminder = divmod(len_deltas, RSYNC_BLOCK_SIZE)
            if len_deltas > 0:
                block_indexes = (
                    data_chunk[offset_size: offset_size + len_offset_str])
                data_chunk = data_chunk[offset_size + len(block_indexes):]
                blocks_offsets = filter(None, block_indexes.split('\00'))
                # Get all the block index offset from
                for block_index in blocks_offsets:
                    while len(data_chunk) < RSYNC_BLOCK_SIZE:
                        try:
                            data_chunk += self.process_restore_data(
                                read_pipe.recv_bytes())
                        except EOFError:
                            LOG.info(
                                "[*] EOF from pipe. Flushing buffer.")
                            data_chunk += self.compressor.flush()
                            break

                    offset = int(block_index) * RSYNC_BLOCK_SIZE
                    fd_curr_file.seek(offset)
                    fd_curr_file.write(data_chunk[:RSYNC_BLOCK_SIZE])
                    data_chunk = data_chunk[RSYNC_BLOCK_SIZE:]

                if reminder:
                    fd_curr_file.write(data_chunk[:reminder])
                    data_chunk = data_chunk[reminder:]

        return data_chunk

    def make_reg_file(
            self, size, file_path, read_pipe, data_chunk,
            flushed, level_id):
        """Create the regular file and write data on it.

        :param size:
        :param file_path:
        :param read_pipe:
        :param data_chunk:
        :param flushed:
        :param level_id:
        :return:
        """

        # File is created. If size is 0, no content is written and the
        # function return

        if level_id == '0000':
            fd_curr_file = open(file_path, 'wb')
            data_chunk = self.write_file(fd_curr_file, size, data_chunk,
                                         read_pipe, flushed)
        elif level_id == '1111':
            fd_curr_file = open(file_path, 'rb+')
            data_chunk = self.write_changes_in_file(fd_curr_file,
                                                    data_chunk, read_pipe)
        fd_curr_file.close()
        return data_chunk

    def set_inode(self, uname, gname, mtime, name):
        """Set the file inode fields according to the provided args.

        :param uname:
        :param gname:
        :param mtime:
        :param name:
        :return:
        """

        try:
            current_uid = pwd.getpwnam(uname).pw_uid
            current_gid = grp.getgrnam(gname).gr_gid
            os.chown(name, current_uid, current_gid)
            os.utime(name, (mtime, mtime))
        except (IOError, OSError):
            try:
                current_uid = pwd.getpwnam(getpass.getuser()).pw_uid
                current_gid = grp.getgrnam(getpass.getuser()).gr_gid
                os.chown(name, current_uid, current_gid)
                os.utime(name, (mtime, mtime))
            except (OSError, IOError) as err:
                raise Exception(err)

    @staticmethod
    def get_file_type(file_mode, fs_path):
        """Extract file type from the the file mode retrieved
        from file abs path

        :param file_mode:
        :param fs_path:
        :return:
        """

        if stat.S_ISREG(file_mode):
            return 'r', ''
        if stat.S_ISDIR(file_mode):
            return 'd', ''
        if stat.S_ISLNK(file_mode):
            return 'l', os.readlink(fs_path)
        if stat.S_ISCHR(file_mode):
            return 'c', ''
        if stat.S_ISBLK(file_mode):
            return 'b', ''
        if stat.S_ISFIFO(file_mode):
            return 'p', ''
        if stat.S_ISSOCK(file_mode):
            return 's', ''

        return 'u', ''

    @staticmethod
    def gen_file_header(file_path, inode_str_struct):
        """Generate file header for rsync binary data file

        :param file_path: file path
        :param inode_str_struct: file binary string including inode data
        :return: chunk of binary data to be processed on the next iteration
        """

        start_of_block = b'\00{}\00'.format(file_path)
        header_size = len(start_of_block) + len(inode_str_struct)
        header_size += len(str(header_size))
        file_header = b'{}{}{}'.format(
            header_size, start_of_block, inode_str_struct)
        len_file_header = len(file_header)

        if header_size != len_file_header:
            file_header_list = file_header.split('\00')
            file_header_list = file_header_list[1:]
            file_header_list.insert(0, str(len_file_header))
            file_header = '\00'.join(file_header_list)
        return file_header

    def make_files(
            self, header_list, restore_abs_path, read_pipe,
            data_chunk, flushed,
            current_backup_level):
        """
        Header list binary structure:

        header_len, file_abs_path, RSYNC_DATA_STRUCT_VERSION, file_mode,
        os_stat.st_uid, os_stat.st_gid, os_stat.st_size,
        mtime, ctime, uname, gname, file_type, linkname, rsync_block_size,

        :param header_list:
        :param restore_abs_path:
        :param read_pipe:
        :param data_chunk:
        :param flushed:
        :param current_backup_level:
        :return:
        """

        # The following commented lines are important for development and
        # troubleshooting, please let it go :)
        # header_len = header_list[0]
        file_path = header_list[1]
        # data_ver = header_list[2]
        file_mode = header_list[3]
        # uid = header_list[4]
        # gid = header_list[5]
        size = header_list[6]
        mtime = header_list[7]
        # ctime = header_list[8]
        uname = header_list[9]
        gname = header_list[10]
        file_type = header_list[11]
        link_name = header_list[12]
        # inumber = header_list[13]
        # nlink = header_list[14]
        devminor = header_list[15]
        devmajor = header_list[16]
        # rsync_block_size = header_list[17]
        level_id = header_list[18]
        rm = header_list[19]

        # Data format conversion
        file_mode = int(file_mode)
        size = int(size)
        mtime = int(mtime)
        file_abs_path = '{0}/{1}'.format(restore_abs_path, file_path)

        if not os.path.isdir(file_abs_path) and os.path.exists(
                file_abs_path) and current_backup_level == 0:
            os.unlink(file_abs_path)

        if not os.path.isdir(file_abs_path) and os.path.exists(
                file_abs_path) and current_backup_level != 0 and rm == '1111':
            os.unlink(file_abs_path)
            return data_chunk

        if file_type in REG_FILE:
            data_chunk = self.make_reg_file(
                size, file_abs_path, read_pipe, data_chunk,
                flushed, level_id)

        elif file_type == 'd':
            try:
                os.makedirs(file_abs_path, file_mode)
            except (OSError, IOError) as error:
                if error.errno != 17:  # E_EXIST
                    LOG.warning(
                        'Directory {0} creation error: {1}'.format(
                            file_abs_path, error))

        elif file_type == 'b':
            file_mode |= stat.S_IFBLK
            try:
                devmajor = int(devmajor)
                devminor = int(devminor)
                new_dev = os.makedev(devmajor, devminor)
                os.mknod(file_abs_path, file_mode, new_dev)
            except (OSError, IOError) as error:
                LOG.warning(
                    'Block file {0} creation error: {1}'.format(
                        file_abs_path, error))

        elif file_type == 'c':
            file_mode |= stat.S_IFCHR
            try:
                devmajor = int(devmajor)
                devminor = int(devminor)
                new_dev = os.makedev(devmajor, devminor)
                os.mknod(file_abs_path, file_mode, new_dev)
            except (OSError, IOError) as error:
                LOG.warning(
                    'Character file {0} creation error: {1}'.format(
                        file_abs_path, error))

        elif file_type == 'p':
            try:
                os.mkfifo(file_abs_path)
            except (OSError, IOError) as error:
                LOG.warning(
                    'FIFO or Pipe file {0} creation error: {1}'.format(
                        file_abs_path, error))

        elif file_type == 'l':
            try:
                os.symlink(link_name, file_abs_path)
            except (OSError, IOError) as error:
                LOG.warning('Link file {0} creation error: {1}'.format(
                    file_abs_path, error))

        if file_type != 'l':
            self.set_inode(uname, gname, mtime, file_abs_path)

        return data_chunk

    def get_file_struct(self, fs_path, new_level=False):
        """Generate file meta data from file abs path.

        Return the meta data as a dict structure and a binary string

        :param fs_path: file abs path
        :param new_level
        :return: file data structure
        """

        # Get file inode information, whether the file is a regular
        # file or a symbolic link
        try:
            os_stat = os.lstat(fs_path)
        except (OSError, IOError) as error:
            raise Exception('[*] Error on file stat: {}'.format(error))

        file_mode = os_stat.st_mode
        # Get file type. If file type is a link it returns also the
        # file pointed by the link
        file_type, lname = self.get_file_type(file_mode, fs_path)

        # If file_type is a socket return False
        if file_type == 's':
            return False, False

        ctime = int(os_stat.st_ctime)
        mtime = int(os_stat.st_mtime)
        uname = pwd.getpwuid(os_stat.st_uid)[0]
        gname = grp.getgrgid(os_stat.st_gid)[0]

        dev = os_stat.st_dev
        inumber = os_stat.st_ino
        nlink = os_stat.st_nlink
        uid = os_stat.st_uid
        gid = os_stat.st_gid
        size = os_stat.st_size
        devmajor = os.major(dev)
        devminor = os.minor(dev)

        level_id = '0000'
        if new_level:
            level_id = '1111'

        # build file meta data as dictionary
        inode_dict = {
            'inode': {
                'inumber': inumber,
                'nlink': nlink,
                'mode': file_mode,
                'uid': uid,
                'gid': gid,
                'size': size,
                'devmajor': devmajor,
                'devminor': devminor,
                'mtime': mtime,
                'ctime': ctime,
                'uname': uname,
                'gname': gname,
                'ftype': file_type,
                'lname': lname,
                'rsync_block_size': RSYNC_BLOCK_SIZE,
                'level_id': level_id,
                'deleted': '0000'
            }
        }

        # build file meta data as binary string
        inode_bin_str = (
            b'{}\00{}\00{}\00{}\00{}'
            b'\00{}\00{}\00{}\00{}\00{}'
            b'\00{}\00{}\00{}\00{}\00{}\00{}\00{}\00{}').format(
            RSYNC_DATA_STRUCT_VERSION, file_mode,
            uid, gid, size, mtime, ctime, uname, gname,
            file_type, lname, inumber, nlink, devminor, devmajor,
            RSYNC_BLOCK_SIZE, level_id, '0000')

        return inode_dict, inode_bin_str

    def gen_struct_for_deleted_files(self, files_meta, old_fs_meta_struct,
                                     rel_path, write_queue):
        files_meta['files'][rel_path] = old_fs_meta_struct['files'][rel_path]
        files_meta['files'][rel_path]['inode']['deleted'] = '1111'
        file_mode = files_meta['files'][rel_path]['inode']['mode']
        uid = files_meta['files'][rel_path]['inode']['uid']
        gid = files_meta['files'][rel_path]['inode']['gid']
        size = files_meta['files'][rel_path]['inode']['size']
        mtime = files_meta['files'][rel_path]['inode']['mtime']
        ctime = files_meta['files'][rel_path]['inode']['ctime']
        uname = files_meta['files'][rel_path]['inode']['uname']
        gname = files_meta['files'][rel_path]['inode']['gname']
        file_type = files_meta['files'][rel_path]['inode']['ftype']
        lname = files_meta['files'][rel_path]['inode']['lname']
        inumber = files_meta['files'][rel_path]['inode']['inumber']
        nlink = files_meta['files'][rel_path]['inode']['nlink']
        devminor = files_meta['files'][rel_path]['inode']['devminor']
        devmajor = files_meta['files'][rel_path]['inode']['devmajor']
        level_id = files_meta['files'][rel_path]['inode']['level_id']

        inode_bin_str = (
            b'{}\00{}\00{}\00{}\00{}'
            b'\00{}\00{}\00{}\00{}\00{}'
            b'\00{}\00{}\00{}\00{}\00{}\00{}\00{}\00{}').format(
            RSYNC_DATA_STRUCT_VERSION, file_mode,
            uid, gid, size, mtime, ctime, uname, gname,
            file_type, lname, inumber, nlink, devminor, devmajor,
            RSYNC_BLOCK_SIZE, level_id, '1111')
        file_header = self.gen_file_header(rel_path, inode_bin_str)
        compr_block = self.process_backup_data(file_header)
        write_queue.put(compr_block)

    def process_file(self, file_path, fs_path, files_meta,
                     old_fs_meta_struct, write_queue):
        rel_path = os.path.relpath(file_path, fs_path)

        new_level = True if self.get_old_file_meta(
            old_fs_meta_struct, rel_path) else False

        inode_dict_struct, inode_str_struct = self.get_file_struct(
            rel_path, new_level)

        if not inode_dict_struct:
            return

        if os.path.isdir(file_path):
            files_meta['directories'][file_path] = inode_dict_struct
            files_meta['meta']['backup_size_on_disk'] += os.path.getsize(
                rel_path)
            file_header = self.gen_file_header(rel_path, inode_str_struct)

            compressed_block = self.process_backup_data(file_header)
            files_meta['meta']['backup_size_compressed'] += len(
                compressed_block)
            write_queue.put(compressed_block)
        else:
            file_metadata = self.compute_incrementals(
                rel_path, inode_str_struct,
                inode_dict_struct, files_meta,
                old_fs_meta_struct, write_queue)

            files_meta.update(file_metadata)

    def get_sign_delta(self, fs_path, manifest_path, write_queue):
        """Compute the file or fs tree path signatures.

        :param fs_path:
        :param manifest_path
        :param write_queue:
        :return:
        """

        files_meta = {
            'files': {},
            'directories': {},
            'meta': {
                'broken_links_tot': '',
                'total_files': '',
                'total_directories': '',
                'backup_size_on_disk': 0,
                'backup_size_uncompressed': 0,
                'backup_size_compressed': 0,
                'platform': sys.platform
            },
            'abs_backup_path': os.getcwd(),
            'broken_links': [],
            'rsync_struct_ver': RSYNC_DATA_STRUCT_VERSION,
            'rsync_block_size': RSYNC_BLOCK_SIZE}

        # Get old file meta structure or an empty dict if not available
        old_fs_meta_struct = self.get_fs_meta_struct(manifest_path)

        if os.path.isdir(fs_path):
            # If given path is a directory, change cwd to path to backup
            os.chdir(fs_path)
            for root, dirs, files in os.walk(fs_path):
                self.process_file(root, fs_path, files_meta,
                                  old_fs_meta_struct, write_queue)

                # Check if exclude is in filename. If it is, log the file
                # exclusion and continue to the next iteration.
                if self.exclude:
                    files = [name for name in files if
                             self.exclude not in name]
                    if files:
                        LOG.warning(
                            ('Excluding file names matching with: '
                             '{}'.format(self.exclude)))

                for name in files:
                    file_path = os.path.join(root, name)
                    self.process_file(file_path, fs_path,
                                      files_meta, old_fs_meta_struct,
                                      write_queue)
        else:
            self.process_file(fs_path, os.getcwd(), files_meta,
                              old_fs_meta_struct, write_queue)
        if old_fs_meta_struct:
            for rel_path in old_fs_meta_struct['files']:
                if not files_meta['files'].get(rel_path):
                    self.gen_struct_for_deleted_files(
                        files_meta, old_fs_meta_struct, rel_path, write_queue)

        # Flush any compressed buffered data
        flushed_data = self.compressor.flush()
        if flushed_data:
            flushed_data = self.process_backup_data(flushed_data,
                                                    do_compress=False)
            files_meta['meta']['backup_size_compressed'] += len(flushed_data)
            write_queue.put(flushed_data)

        # General metrics to be uploaded to the API and/or media storage
        files_meta['meta']['broken_links_tot'] = len(
            files_meta['broken_links'])
        files_meta['meta']['total_files'] = len(files_meta['files'])
        files_meta['meta']['total_directories'] = len(
            files_meta['directories'])
        files_meta['meta']['rsync_data_struct_ver'] = RSYNC_DATA_STRUCT_VERSION
        LOG.info("Backup session metrics: {0}".format(
            files_meta['meta']))

        # Compress meta data file
        # Write meta data to disk as JSON
        compressed_json_meta = compress.one_shot_compress(
            self.compression_algo, json.dumps(files_meta))
        with open(manifest_path, 'wb') as manifest_file:
            manifest_file.write(compressed_json_meta)

        # Put False on the queue so it will be terminated on the other side:
        write_queue.put(False)

    def get_fs_meta_struct(self, fs_meta_path):
        fs_meta_struct = {}

        if os.path.isfile(fs_meta_path):
            with open(fs_meta_path) as meta_file:
                fs_meta_struct = json.loads(
                    compress.one_shot_decompress(self.compression_algo,
                                                 meta_file.read()))

        return fs_meta_struct

    @staticmethod
    def is_reg_file(file_type):
        if file_type in REG_FILE:
            return True
        return False

    @staticmethod
    def get_old_file_meta(old_fs_meta_struct, rel_path):
        if old_fs_meta_struct:
            return old_fs_meta_struct['files'].get(rel_path)
        return None

    @staticmethod
    def compute_checksums(rel_path, files_meta, reg_file=True):
        # Files type where the file content can be backed up
        if reg_file:
            with open(rel_path, 'rb') as file_path_fd:
                files_meta['files'][rel_path].update(
                    {'signature': pyrsync.blockchecksums(
                        file_path_fd, RSYNC_BLOCK_SIZE)})
        else:
            # Stat the file to be sure it's not a broken link
            if os.path.lexists(rel_path):
                if not os.path.exists(rel_path):
                    raise IOError

            files_meta['files'][rel_path].update(
                {'signature': [[], []]})

        return files_meta

    def compute_incrementals(
            self, rel_path, inode_str_struct,
            inode_dict_struct, files_meta,
            old_fs_meta_struct, write_queue, deleted=False):

        file_header = self.gen_file_header(
            rel_path, inode_str_struct)

        # File size is header size + null bytes + file size
        file_size = len(file_header)
        reg_file_type = self.is_reg_file(inode_dict_struct['inode']['ftype'])
        try:
            files_meta['files'][rel_path] = inode_dict_struct

            # As pyrsync functions run thorough the file descriptor
            # we need to set the set the pointer to the fd to 0.
            # Backup file data content only if the file type is
            # regular or unknown
            old_file_meta = self.get_old_file_meta(old_fs_meta_struct,
                                                   rel_path)
            if old_file_meta:
                if self.is_file_modified(old_file_meta,
                                         files_meta['files'][rel_path]):
                    # If old_fs_path is provided, it checks
                    # if old file mtime or ctime are different then
                    # the new ones
                    files_meta = self.compute_checksums(rel_path, files_meta,
                                                        reg_file=reg_file_type)

                    compressed_block = self.process_backup_data(file_header)
                    write_queue.put(compressed_block)

                    if reg_file_type:
                        file_path_fd = open(rel_path)
                        for data_block in self.rsync_gen_delta(
                                file_path_fd, old_file_meta):

                            compressed_block = self.process_backup_data(
                                data_block)
                            write_queue.put(compressed_block)
                        file_path_fd.close()
                else:
                    files_meta['files'][rel_path].update(
                        {'signature': old_file_meta['signature']})

            else:
                files_meta = self.compute_checksums(
                    rel_path, files_meta,
                    reg_file=reg_file_type)

                compressed_block = self.process_backup_data(file_header)
                write_queue.put(compressed_block)
                if reg_file_type:
                    file_path_fd = open(rel_path)
                    data_block = file_path_fd.read(
                        RSYNC_BLOCK_BUFF_SIZE)
                    while data_block:
                        compressed_block = self.process_backup_data(data_block)
                        write_queue.put(compressed_block)
                        data_block = file_path_fd.read(
                            RSYNC_BLOCK_BUFF_SIZE)

                    file_path_fd.close()
            files_meta['files'][rel_path]['file_data_len'] = file_size
        except (IOError, OSError) as error:
            LOG.warning('IO or OS Error: {}'.format(error))
            if os.path.lexists(rel_path):
                LOG.warning(
                    'Broken link at: {}'.format(rel_path))
                files_meta['broken_links'].append(rel_path)

        return files_meta



"""
(c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Freezer general utils functions
"""
import os
import subprocess

from oslo_log import log

from freezer.engine import engine
from freezer.engine.tar import tar_builders
from freezer.utils import winutils

LOG = log.getLogger(__name__)


class TarEngine(engine.BackupEngine):

    def __init__(
            self, compression, symlinks, exclude, storage,
            max_segment_size, encrypt_key=None,
            dry_run=False):
        """
            :type storage: freezer.storage.base.Storage
        :return:
        """
        self.compression_algo = compression
        self.encrypt_pass_file = encrypt_key
        self.dereference_symlink = symlinks
        self.exclude = exclude
        self.storage = storage
        self.is_windows = winutils.is_windows()
        self.dry_run = dry_run
        self.max_segment_size = max_segment_size
        super(TarEngine, self).__init__(storage=storage)

    @property
    def name(self):
        return "tar"

    def metadata(self, *args):
        return {
            "engine_name": self.name,
            "compression": self.compression_algo,
            # the encrypt_pass_file might be key content so we need to covert
            # to boolean
            "encryption": bool(self.encrypt_pass_file)
        }

    def backup_data(self, backup_resource, manifest_path):
        LOG.info("Starting Tar engine backup stream")
        tar_command = tar_builders.TarCommandBuilder(
            backup_resource, self.compression_algo, self.is_windows)
        if self.encrypt_pass_file:
            tar_command.set_encryption(self.encrypt_pass_file)
        if self.dereference_symlink:
            tar_command.set_dereference(self.dereference_symlink)
        tar_command.set_exclude(self.exclude)
        tar_command.set_listed_incremental(manifest_path)

        command = tar_command.build()

        LOG.info("Execution command: \n{}".format(command))

        tar_process = subprocess.Popen(
            command, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            shell=True, executable='/bin/bash')
        read_pipe = tar_process.stdout
        tar_chunk = read_pipe.read(self.max_segment_size)
        while tar_chunk:
            yield tar_chunk
            tar_chunk = read_pipe.read(self.max_segment_size)

        self.check_process_output(tar_process, 'Backup')

        LOG.info("Tar engine stream completed")

    def restore_level(self, restore_resource, read_pipe, backup, except_queue):
        """
        Restore the provided file into backup_opt_dict.restore_abs_path
        Decrypt the file if backup_opt_dict.encrypt_pass_file key is provided

        :param restore_path:
        :param read_pipe:
        :type backup: freezer.storage.base.Backup
        :param backup:
        """

        try:

            metadata = backup.metadata()
            if (not self.encrypt_pass_file and
                    metadata.get("encryption", False)):
                raise Exception("Cannot restore encrypted backup without key")

            tar_command = tar_builders.TarCommandRestoreBuilder(
                restore_resource,
                metadata.get('compression', self.compression_algo),
                self.is_windows)

            if self.encrypt_pass_file:
                tar_command.set_encryption(self.encrypt_pass_file)

            if self.dry_run:
                tar_command.set_dry_run()

            command = tar_command.build()

            if winutils.is_windows():
                # on windows, chdir to restore path.
                os.chdir(restore_resource)

            tar_process = subprocess.Popen(
                command, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE, shell=True, executable='/bin/bash')
            # Start loop reading the pipe and pass the data to the tar
            # std input. If EOFError exception is raised, the loop end
            # the std err will be checked for errors.
            try:
                while True:
                    tar_process.stdin.write(read_pipe.recv_bytes())
            except EOFError:
                LOG.info('Pipe closed as EOF reached. '
                         'Data transmitted successfully')
            finally:
                self.check_process_output(tar_process, 'Restore')

        except Exception as e:
            LOG.exception(e)
            except_queue.put(e)
            raise

    @staticmethod
    def check_process_output(process, function):

        """Check process stderr and return code to determine if error occurred.

        Check process stderr for any non-empty value.
        Check process return code for any non-zero value.
        Log error message and raise an exception if error occurred.

        :param process: the multiprocessing process to check
        :param function: a string: ('Restore' | 'Backup') for error message
        :returns: None -- Do nothing and return None if no error occurred
        :raises: Exception -- Raise Exception if error occurred
        """
        tar_err = process.communicate()[1]

        if tar_err:
            LOG.error('{0} error: {1}'.format(function, tar_err))

        if process.returncode:
            LOG.error('{0} return code is not 0'
                      .format(process.returncode))
            raise Exception('{0} process failed with return code: {1}'
                            .format(function, process.returncode))



"""
Copyright 2015 Hewlett-Packard

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import datetime
import json
import os
import subprocess
import tempfile
import time

from freezer.utils import utils
from oslo_config import cfg
from oslo_log import log
from six.moves import configparser


CONF = cfg.CONF
LOG = log.getLogger(__name__)


class StopState(object):

    @staticmethod
    def stop(job, doc):
        job.job_doc = doc
        job.event = Job.NO_EVENT
        job.job_doc_status = Job.STOP_STATUS
        job.scheduler.update_job(job.id, job.job_doc)
        return Job.NO_EVENT

    @staticmethod
    def abort(job, doc):
        return StopState.stop(job, doc)

    @staticmethod
    def start(job, doc):
        job.job_doc = doc
        job.event = Job.NO_EVENT
        job.job_doc_status = Job.STOP_STATUS
        job.schedule()
        job.scheduler.update_job(job.id, job.job_doc)
        return Job.NO_EVENT

    @staticmethod
    def remove(job):
        job.unschedule()
        job.job_doc_status = Job.REMOVED_STATUS
        return Job.NO_EVENT


class ScheduledState(object):

    @staticmethod
    def stop(job, doc):
        job.unschedule()
        job.scheduler.update_job(job.id, job.job_doc)
        return Job.STOP_EVENT

    @staticmethod
    def abort(job, doc):
        return StopState.stop(job, doc)

    @staticmethod
    def start(job, doc):
        job.event = Job.NO_EVENT
        job.scheduler.update_job(job.id, job.job_doc)
        return Job.NO_EVENT

    @staticmethod
    def remove(job):
        job.unschedule()
        job.job_doc_status = Job.REMOVED_STATUS
        return Job.NO_EVENT


class RunningState(object):

    @staticmethod
    def stop(job, doc):
        job.event = Job.STOP_EVENT
        return Job.NO_EVENT

    @staticmethod
    def abort(job, doc):
        job.event = Job.ABORT_EVENT
        job.scheduler.update_job(job.id, job.job_doc)
        return Job.ABORTED_RESULT

    @staticmethod
    def start(job, doc):
        job.event = Job.NO_EVENT
        job.scheduler.update_job(job.id, job.job_doc)
        return Job.NO_EVENT

    @staticmethod
    def remove(job):
        job.event = Job.REMOVE_EVENT
        return Job.NO_EVENT


class Job(object):

    NO_EVENT = ''
    STOP_EVENT = 'stop'
    START_EVENT = 'start'
    ABORT_EVENT = 'abort'
    REMOVE_EVENT = 'remove'

    STOP_STATUS = 'stop'
    SCHEDULED_STATUS = 'scheduled'
    RUNNING_STATUS = 'running'
    REMOVED_STATUS = 'removed'
    COMPLETED_STATUS = 'completed'

    FAIL_RESULT = 'fail'
    SUCCESS_RESULT = 'success'
    ABORTED_RESULT = 'aborted'

    TIME_NULL = -1

    @staticmethod
    def create(scheduler, executable, job_doc):
        job = Job(scheduler, executable, job_doc)
        if job.job_doc_status in ['running', 'scheduled']:
            LOG.warning('Resetting {0} status from job {1}'
                        .format(job.job_doc_status, job.id))
        if job.job_doc_status == 'stop' and not job.event:
            LOG.info('Job {0} was stopped.'.format(job.id))
            job.event = Job.STOP_EVENT
        elif not job.event:
            LOG.info('Autostart Job {0}'.format(job.id))
            job.event = Job.START_EVENT
        return job

    def __init__(self, scheduler, executable, job_doc):
        self.scheduler = scheduler
        self.executable = executable
        self.job_doc = job_doc
        self.process = None
        self.state = StopState

    def remove(self):
        with self.scheduler.lock:
            # delegate to state object
            LOG.info('REMOVE job {0}'.format(self.id))
            self.state.remove(self)

    @property
    def id(self):
        return self.job_doc['job_id']

    @property
    def session_id(self):
        return self.job_doc.get('session_id', '')

    @session_id.setter
    def session_id(self, value):
        self.job_doc['session_id'] = value

    @property
    def session_tag(self):
        return self.job_doc.get('session_tag', 0)

    @session_tag.setter
    def session_tag(self, value):
        self.job_doc['session_tag'] = value

    @property
    def event(self):
        return self.job_doc['job_schedule'].get('event', '')

    @event.setter
    def event(self, value):
        self.job_doc['job_schedule']['event'] = value

    @property
    def job_doc_status(self):
        return self.job_doc['job_schedule'].get('status', '')

    @job_doc_status.setter
    def job_doc_status(self, value):
        self.job_doc['job_schedule']['status'] = value

    @property
    def result(self):
        return self.job_doc['job_schedule'].get('result', '')

    @result.setter
    def result(self, value):
        self.job_doc['job_schedule']['result'] = value

    def can_be_removed(self):
        return self.job_doc_status == Job.REMOVED_STATUS

    @staticmethod
    def save_action_to_file(action, f):
        parser = configparser.ConfigParser()
        parser.add_section('action')
        for action_k, action_v in action.items():
            parser.set('action', action_k, action_v)
        parser.write(f)
        f.seek(0)

    @property
    def schedule_date(self):
        return self.job_doc['job_schedule'].get('schedule_date', '')

    @property
    def schedule_interval(self):
        return self.job_doc['job_schedule'].get('schedule_interval', '')

    @property
    def schedule_start_date(self):
        return self.job_doc['job_schedule'].get('schedule_start_date', '')

    @property
    def schedule_end_date(self):
        return self.job_doc['job_schedule'].get('schedule_end_date', '')

    @property
    def schedule_cron_fields(self):
        cron_fields = ['year', 'month', 'day', 'week', 'day_of_week',
                       'hour', 'minute', 'second']
        cron_schedule = {}
        for cron in self.job_doc['job_schedule'].keys():
            if cron.startswith('schedule_'):
                cron_key = cron.split('_', 1)[1]
                cron_schedule.update({
                    cron_key: self.job_doc['job_schedule'][cron]})
        return {key: value
                for key, value in cron_schedule.items()
                if key in cron_fields}

    @property
    def scheduled(self):
        return self.scheduler.is_scheduled(self.id)

    def get_schedule_args(self):
        def get_start_date(date):
            # start_date format "%Y-%m-%dT%H:%M:%S"
            now = datetime.datetime.now()
            start_date = now + datetime.timedelta(0, 2, 0)
            if (utils.date_to_timestamp(date) >
                    utils.date_to_timestamp(now.isoformat().split('.')[0])):
                start_date = datetime.datetime.strptime(
                    date, "%Y-%m-%dT%H:%M:%S")
            return start_date

        def get_end_date(start, end):
            # start end format "%Y-%m-%dT%H:%M:%S"
            end_date = datetime.datetime.strptime(end, "%Y-%m-%dT%H:%M:%S")
            if (utils.date_to_timestamp(start) > utils.date_to_timestamp(end)):
                end_date = None
            return end_date
        kwargs_date = {}
        if self.schedule_start_date:
            kwargs_date.update({
                'start_date': get_start_date(self.schedule_start_date)
            })
            if self.schedule_end_date:
                end_date = get_end_date(self.schedule_start_date,
                                        self.schedule_end_date)
                kwargs_date.update({
                    'end_date': end_date
                })

        if self.schedule_date:
            return {'trigger': 'date',
                    'run_date': self.schedule_date}
        elif self.schedule_interval:
            kwargs = {'trigger': 'interval'}
            kwargs.update(kwargs_date)
            if self.schedule_interval == 'continuous':
                kwargs.update({'seconds': 1})
            else:
                val, unit = self.schedule_interval.split(' ')
                kwargs.update({unit: int(val)})
            return kwargs
        elif self.schedule_cron_fields:
            kwargs = {'trigger': 'cron'}
            kwargs.update(kwargs_date)
            cron_fields = self.schedule_cron_fields
            kwargs.update(cron_fields)
            return kwargs
        else:
            # no scheduling information, schedule to start within a few seconds
            return {'trigger': 'date',
                    'run_date': datetime.datetime.now() +
                    datetime.timedelta(0, 2, 0)}

    def process_event(self, job_doc):
        with self.scheduler.lock:
            next_event = job_doc['job_schedule'].get('event', '')
            while next_event:
                if next_event == Job.STOP_EVENT:
                    if isinstance(self.state(), StopState):
                        LOG.info('JOB {0} event: STOP'.format(self.id))
                    next_event = self.state.stop(self, job_doc)
                elif next_event == Job.START_EVENT:
                    LOG.info('JOB {0} event: START'.format(self.id))
                    next_event = self.state.start(self, job_doc)
                elif next_event == Job.ABORT_EVENT:
                    LOG.info('JOB {0} event: ABORT'.format(self.id))
                    next_event = self.state.abort(self, job_doc)
                elif next_event == Job.ABORTED_RESULT:
                    LOG.info('JOB {0} aborted.'.format(self.id))
                    break

    def upload_metadata(self, metadata_string):
        try:
            metadata = json.loads(metadata_string)
            if metadata:
                metadata['job_id'] = self.id
                self.scheduler.upload_metadata(metadata)
                LOG.info("Job {0}, freezer action metadata uploaded"
                         .format(self.id))
        except Exception as e:
            LOG.error('metrics upload error: {0}'.format(e))

    def execute_job_action(self, job_action):
        max_tries = (job_action.get('max_retries', 0) + 1)
        tries = max_tries
        freezer_action = job_action.get('freezer_action', {})
        max_retries_interval = job_action.get('max_retries_interval', 60)
        action_name = freezer_action.get('action', '')
        while tries:
            with tempfile.NamedTemporaryFile(delete=False) as config_file:
                self.save_action_to_file(freezer_action, config_file)
                config_file_name = config_file.name
                freezer_command = '{0} --metadata-out - --config {1}'.\
                    format(self.executable, config_file.name)
                self.process = subprocess.Popen(freezer_command.split(),
                                                stdout=subprocess.PIPE,
                                                stderr=subprocess.PIPE,
                                                env=os.environ.copy())

                # store the pid for this process in the api
                try:
                    self.job_doc['job_schedule']['current_pid'] = \
                        self.process.pid

                    self.scheduler.update_job(self.job_doc['job_id'],
                                              self.job_doc)
                except Exception as error:
                    LOG.error("Error saving the process id {}".format(error))

                output, error = self.process.communicate()
                # ensure the tempfile gets deleted
                utils.delete_file(config_file_name)

            if error:
                LOG.error("Freezer client error: {0}".format(error))
            elif output:
                self.upload_metadata(output)

            if self.process.returncode == -15:
                # This means the job action was aborted by the scheduler
                LOG.warning('Freezer-agent was killed by the scheduler. '
                            'Cleanup should be done manually: container, '
                            'mountpoint and lvm snapshots.')
                return Job.ABORTED_RESULT

            elif self.process.returncode:
                # ERROR
                tries -= 1
                if tries:
                    LOG.warning('Job {0} failed {1} action,'
                                ' retrying in {2} seconds'
                                .format(self.id, action_name,
                                        max_retries_interval))
                    time.sleep(max_retries_interval)
            else:
                # SUCCESS
                LOG.info('Job {0} action {1}'
                         ' returned success exit code'.
                         format(self.id, action_name))
                return Job.SUCCESS_RESULT
        LOG.error('Job {0} action {1} failed after {2} tries'
                  .format(self.id, action_name, max_tries))

        return Job.FAIL_RESULT

    def contains_exec(self):
        jobs = self.job_doc.get('job_actions')
        for job in jobs:
            freezer_action = job.get('freezer_action')
            action = freezer_action.get('action')
            if action == 'exec':
                return True
        return False

    def update_job_schedule_doc(self, **kwargs):
        """
        Updates the job_schedule section of the job doc with the
        provided keyword args. No checks about accepted key/values
        are being made here since they may vary in the future.

        :param kwargs: keyword args to add
        :return: None
        """
        job_schedule = self.job_doc['job_schedule']
        job_schedule.update(kwargs)

    def execute(self):
        result = Job.SUCCESS_RESULT
        with self.scheduler.lock:
            LOG.info('job {0} running'.format(self.id))
            self.state = RunningState
            self.update_job_schedule_doc(status=Job.RUNNING_STATUS,
                                         result="",
                                         time_started=int(time.time()),
                                         time_ended=Job.TIME_NULL)
            self.scheduler.update_job_schedule(
                self.id,
                self.job_doc['job_schedule'])

        self.start_session()
        # if the job contains exec action and the scheduler passes the
        # parameter --disable-exec job execution should fail
        if self.contains_exec() and CONF.disable_exec:
            LOG.info("Job {0} failed because it contains exec action "
                     "and exec actions are disabled by scheduler"
                     .format(self.id))
            self.result = Job.FAIL_RESULT
            self.finish()
            return

        for job_action in self.job_doc.get('job_actions', []):
            if job_action.get('mandatory', False) or\
                    (result == Job.SUCCESS_RESULT):

                action_result = self.execute_job_action(job_action)
                if action_result == Job.FAIL_RESULT:
                    result = Job.FAIL_RESULT

                if action_result == Job.ABORTED_RESULT:
                    result = Job.ABORTED_RESULT
            else:
                freezer_action = job_action.get('freezer_action', {})
                action_name = freezer_action.get('action', '')
                LOG.warning("skipping {0} action".
                            format(action_name))
        self.result = result
        self.finish()

    def finish(self):
        self.update_job_schedule_doc(time_ended=int(time.time()))
        self.end_session(self.result)
        with self.scheduler.lock:
            if self.event == Job.REMOVE_EVENT:
                self.unschedule()
                self.job_doc_status = Job.REMOVED_STATUS
                return

            if not self.scheduled:
                self.job_doc_status = Job.COMPLETED_STATUS
                self.state = StopState
                self.scheduler.update_job(self.id, self.job_doc)
                return

            if self.event in [Job.STOP_EVENT, Job.ABORT_EVENT]:
                self.unschedule()
                self.job_doc_status = Job.COMPLETED_STATUS
                self.scheduler.update_job(self.id, self.job_doc)
            else:
                self.job_doc_status = Job.SCHEDULED_STATUS
                self.state = ScheduledState
                self.scheduler.update_job_schedule(
                    self.id,
                    self.job_doc['job_schedule'])

    def start_session(self):
        if not self.session_id:
            return
        retry = 5
        while retry:
            try:
                resp = self.scheduler.start_session(self.session_id,
                                                    self.id,
                                                    self.session_tag)
                if resp['result'] == 'success':
                    self.session_tag = resp['session_tag']
                    return
            except Exception as e:
                LOG.error('Error while starting session {0}. {1}'.
                          format(self.session_id, e))
            LOG.warning('Retrying to start session {0}'.
                        format(self.session_id))
            retry -= 1
        LOG.error('Unable to start session {0}'.format(self.session_id))

    def end_session(self, result):
        if not self.session_id:
            return
        retry = 5
        while retry:
            try:
                resp = self.scheduler.end_session(self.session_id,
                                                  self.id,
                                                  self.session_tag,
                                                  result)
                if resp['result'] == 'success':
                    return
            except Exception as e:
                LOG.error('Error while ending session {0}. {1}'.
                          format(self.session_id, e))
            LOG.warning('Retrying to end session {0}'.
                        format(self.session_id))
            retry -= 1
        LOG.error('Unable to end session {0}'.format(self.session_id))

    def schedule(self):
        try:
            kwargs = self.get_schedule_args()
            self.scheduler.add_job(self.execute, id=self.id,
                                   executor='threadpool',
                                   misfire_grace_time=3600, **kwargs)
        except Exception as e:
            LOG.error("Unable to schedule job {0}: {1}".
                      format(self.id, e))

        LOG.info('scheduler job with parameters {0}'.format(kwargs))

        if self.scheduled:
            self.job_doc_status = Job.SCHEDULED_STATUS
            self.state = ScheduledState
        else:
            # job not scheduled or already started and waiting for lock
            self.job_doc_status = Job.COMPLETED_STATUS
            self.state = StopState

    def unschedule(self):
        try:
            # already executing job are not present in the apscheduler list
            self.scheduler.remove_job(job_id=self.id)
        except Exception:
            pass
        self.event = Job.NO_EVENT
        self.job_doc_status = Job.STOP_STATUS
        self.state = StopState

    def terminate(self):
        if self.process:
            self.process.terminate()

    def kill(self):
        if self.process:
            self.process.kill()



# (c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import subprocess


def execute(cmd):
    """
    Split a command specified as function arguments into separate sub commands
    executed separately
    """
    cmds = cmd.split('|')
    nb_process = len(cmds)
    index = 1
    process = None
    for sub_cmd in cmds:
        is_last_process = (index == nb_process)
        process = popen_call(sub_cmd.split(' '), process, is_last_process)
        index += 1


def popen_call(sub_cmd, input, is_last_process):
    """
    Execute a command specified as function arguments using the given input
    """
    if not input:
        process = subprocess.Popen(sub_cmd,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE, shell=False)
    else:
        process = subprocess.Popen(sub_cmd,
                                   stdout=subprocess.PIPE, stdin=input.stdout,
                                   stderr=subprocess.PIPE, shell=False)
        input.stdout.close()
    if (is_last_process):
        process.communicate()[0]
        rc = process.returncode
        if rc != 0:
            raise Exception('Error: while executing script '
                            '%s return code was %d instead of 0'
                            % (' '.join(sub_cmd), rc))
    return process



"""
(c) Copyright 2014,2015 Hewlett-Packard Development Company, L.P.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Freezer restore modes related functions
"""

import json
import os
import time

from oslo_config import cfg
from oslo_log import log

from freezer.utils import utils

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class RestoreOs(object):
    def __init__(self, client_manager, container, storage):
        self.client_manager = client_manager
        self.container = container
        self.storage = storage

    def _get_backups(self, path, restore_from_timestamp):
        """
        :param path:
        :type path: str
        :param restore_from_timestamp:
        :type restore_from_timestamp: int
        :return:
        """
        if self.storage.type == "swift":
            swift = self.client_manager.get_swift()
            path = "{0}_segments/{1}/".format(self.container, path)
            info, backups = swift.get_container(self.container, prefix=path)
            backups = sorted(
                map(lambda x: int(x["name"].rsplit("/", 1)[-1]), backups))
        elif self.storage.type == "s3":
            bucket_name, path = self.get_storage_info(path)
            backups = self.storage.list_all_objects(
                bucket_name=bucket_name,
                prefix=path
            )
            backups = sorted(
                map(lambda x: int(x['Key'].split("/", 2)[1]), backups))
        elif self.storage.type == "local":
            path = "{0}/{1}".format(self.container, path)
            backups = os.listdir(os.path.abspath(path))
        elif self.storage.type == "ssh":
            path = "{0}/{1}".format(self.container, path)
            backups = self.storage.listdir(path)
        else:
            msg = ("{} storage type is not supported at the moment."
                   " Try local, swift or ssh".format(self.storage.type))
            print(msg)
            raise BaseException(msg)
        backups = list(filter(lambda x: x >= restore_from_timestamp, backups))
        if not backups:
            msg = "Cannot find backups for path: %s" % path
            LOG.error(msg)
            raise BaseException(msg)
        return backups[-1]

    def get_storage_info(self, path):
        storage_info = self.container.split('/', 1)
        container_name = storage_info[0]
        object_prefix = storage_info[1] if '/' in self.container else ''
        if object_prefix != '':
            path = "{0}/{1}".format(object_prefix, path)
        return container_name, path

    def _create_image(self, path, restore_from_timestamp):
        """
        :param path:
        :param restore_from_timestamp:
        :type restore_from_timestamp: int
        :return:
        """
        backup = self._get_backups(path, restore_from_timestamp)
        if self.storage.type == 'swift':
            swift = self.client_manager.get_swift()
            path = "{0}_segments/{1}/{2}".format(self.container, path, backup)
            stream = swift.get_object(self.container,
                                      "{}/{}".format(path, backup),
                                      resp_chunk_size=10000000)
            length = int(stream[0]["x-object-meta-length"])
            data = utils.ReSizeStream(stream[1], length, 1)
            info = stream[0]
            image = self.client_manager.create_image(
                name="restore_{}".format(path),
                container_format="bare",
                disk_format="raw",
                data=data)
            return info, image
        elif self.storage.type == 's3':
            if self.storage.get_object_prefix() != '':
                base_path = "{0}/{1}/{2}".format(
                    self.storage.get_object_prefix(),
                    path,
                    backup
                )
            else:
                base_path = "{0}/{1}".format(path, backup)
            image_file = "{0}/{1}".format(base_path, path)
            s3_object = self.storage.get_object(
                bucket_name=self.storage.get_bucket_name(),
                key=image_file
            )
            stream = utils.S3ResponseStream(data=s3_object['Body'],
                                            chunk_size=10000000)
            data = utils.ReSizeStream(
                stream,
                s3_object['ContentLength'],
                1
            )
            metadata = "{0}/metadata".format(base_path)
            metadata_object = self.storage.get_object(
                bucket_name=self.storage.get_bucket_name(),
                key=metadata
            )
            info = json.load(metadata_object['Body'])

            image = self.client_manager.create_image(
                name="restore_{}".format(path),
                container_format="bare",
                disk_format="raw",
                data=data)
            return info, image
        elif self.storage.type == 'local':
            image_file = "{0}/{1}/{2}/{3}".format(self.container, path,
                                                  backup, path)
            metadata_file = "{0}/{1}/{2}/metadata".format(self.container,
                                                          path, backup)
            try:
                data = open(image_file, 'rb')
            except Exception:
                msg = "Failed to open image file {}".format(image_file)
                LOG.error(msg)
                raise BaseException(msg)
            info = json.load(file(metadata_file))
            image = self.client_manager.create_image(
                name="restore_{}".format(path),
                container_format="bare",
                disk_format="raw",
                data=data)
            return info, image
        elif self.storage.type == 'ssh':
            image_file = "{0}/{1}/{2}/{3}".format(self.container, path,
                                                  backup, path)
            metadata_file = "{0}/{1}/{2}/metadata".format(self.container,
                                                          path, backup)
            try:
                data = self.storage.open(image_file, 'rb')
            except Exception:
                msg = "Failed to open remote image file {}".format(image_file)
                LOG.error(msg)
                raise BaseException(msg)
            info = json.loads(self.storage.read_metadata_file(metadata_file))
            image = self.client_manager.create_image(
                name="restore_{}".format(path),
                container_format="bare",
                disk_format="raw",
                data=data)
            return info, image
        else:
            return {}

    def restore_cinder(self, volume_id=None,
                       backup_id=None,
                       restore_from_timestamp=None):
        """
        Restoring cinder backup using
        :param volume_id:
        :param backup_id:
        :param restore_from_timestamp:
        :return:
        """
        cinder = self.client_manager.get_cinder()
        search_opts = {
            'volume_id': volume_id,
            'status': 'available',
        }
        if not backup_id:
            backups = cinder.backups.list(search_opts=search_opts)

            def get_backups_from_timestamp(backups, restore_from_timestamp):
                for backup in backups:
                    backup_created_date = backup.created_at.split('.')[0]
                    backup_created_timestamp = (
                        utils.date_to_timestamp(backup_created_date))
                    if backup_created_timestamp >= restore_from_timestamp:
                        yield backup

            backups_filter = get_backups_from_timestamp(backups,
                                                        restore_from_timestamp)
            if not backups_filter:
                LOG.warning("no available backups for cinder volume,"
                            "restore newest backup")
                backup = max(backups, key=lambda x: x.created_at)
            else:
                backup = min(backups_filter, key=lambda x: x.created_at)
            backup_id = backup.id
        cinder.restores.restore(backup_id=backup_id)

    def restore_cinder_by_glance(self, volume_id, restore_from_timestamp):
        """
        1) Define swift directory
        2) Download and upload to glance
        3) Create volume from glance
        4) Delete
        :param restore_from_timestamp:
        :type restore_from_timestamp: int
        :param volume_id - id of attached cinder volume
        """
        (info, image) = self._create_image(volume_id, restore_from_timestamp)
        length = int(info["x-object-meta-length"])
        gb = 1073741824
        size = length / gb
        if length % gb > 0:
            size += 1
        LOG.info("Creating volume from image")
        cinder_client = self.client_manager.get_cinder()
        volume = cinder_client.volumes.create(
            size,
            imageRef=image.id,
            name=info.get('volume_name',
                          CONF.get('backup_name',
                                   CONF.get('cinder_vol_id', None)
                                   )
                          )
        )
        while volume.status != "available":
            try:
                LOG.info("Volume copy status: " + volume.status)
                volume = cinder_client.volumes.get(volume.id)
                if volume.status == "error":
                    raise Exception("Volume copy status: error")
                time.sleep(5)
            except Exception as e:
                LOG.exception(e)
                if volume.status != "error":
                    LOG.warn("Exception getting volume status")

        LOG.info("Deleting temporary image {}".format(image.id))
        self.client_manager.get_glance().images.delete(image.id)



"""
(c) Copyright 2016 Hewlett-Packard Development Enterprise, L.P.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from freezer.openstack import osclients
from freezer.utils import config
from oslo_config import cfg
from oslo_log import log

CONF = cfg.CONF
LOG = log.getLogger(__name__)


def parse_osrc(file_name):
    with open(file_name, 'r') as osrc_file:
        return config.osrc_parse(osrc_file.read())


def get_client_manager(backup_args):
    if "osrc" in backup_args:
        options = osclients.OpenstackOpts.create_from_dict(
            parse_osrc(backup_args['osrc']))
    else:
        options = osclients.OpenstackOpts.create_from_env().get_opts_dicts()
    if backup_args['project_id']:
        options['project_name'] = None
        options['project_id'] = backup_args['project_id']
    client_manager = osclients.OSClientManager(
        auth_url=options.pop('auth_url', None),
        auth_method=options.pop('auth_method', 'password'),
        dry_run=backup_args.get('dry_run', None),
        **options
    )
    return client_manager
